from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix

from fipy.meshes.mesh1D import Mesh1D
from fipy.meshes.mesh2D import Mesh2D
from fipy.meshes.mesh import Mesh

from fipy.meshes.topologies.abstractTopology import _AbstractTopology

class _GridTopology(_AbstractTopology):

    @property
    def _isOrthogonal(self):
        return True

class _Grid1DTopology(_GridTopology):

    _concatenatedClass = Mesh1D

    @property
    def _globalNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1] for mesh A

        ```
            A        B
        ------------------
        | 0 | 1 || 2 | 3 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """

        return numerix.arange(self.mesh.offset + self.mesh.overlap['left'],
                              self.mesh.offset + self.mesh.nx - self.mesh.overlap['right'])

    @property
    def _globalOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2] for mesh A

        ```
            A        B
        ------------------
        | 0 | 1 || 2 | 3 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.offset, self.mesh.offset + self.mesh.nx)

    @property
    def _localNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1] for mesh A

        ```
            A        B
        ------------------
        | 0 | 1 || 1 | 2 |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.overlap['left'],
                              self.mesh.nx - self.mesh.overlap['right'])

    @property
    def _localOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2] for mesh A

        ```
            A        B
        ------------------
        | 0 | 1 || 2 |   |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(0, self.mesh.nx)

    @property
    def _globalNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2] for mesh A

        ```
            A    ||   B
        ------------------
        0   1    2   3   4
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.offset + self.mesh.overlap['left'],
                              self.mesh.offset + self.mesh.numberOfFaces - self.mesh.overlap['right'])

    @property
    def _globalOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3] for mesh A

        ```
            A    ||   B
        ------------------
        0   1    2   3   4
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.offset, self.mesh.offset + self.mesh.numberOfFaces)

    @property
    def _localNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2] for mesh A

        ```
            A    ||   B
        ------------------
        0   1   2/1  2   3
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.overlap['left'],
                              self.mesh.numberOfFaces - self.mesh.overlap['right'])

    @property
    def _localOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3] for mesh A

        ```
            A   ||   B
        ------------------
        0   1   2   3    |
        ------------------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(0, self.mesh.numberOfFaces)

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell of grid"""
        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)
        cellTopology[:] = self._elementTopology["line"]

        return cellTopology

class _Grid2DTopology(_GridTopology):

    _concatenatedClass = Mesh2D

    @property
    def _globalNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1] for mesh A

        ```
        ---------
        | 4 | 5 |
        ---------  B
        | 2 | 3 |
        =========
        | 0 | 1 |  A
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange((self.mesh.offset[1] + self.mesh.overlap['bottom']) * self.mesh.nx,
                              (self.mesh.offset[1] + self.mesh.ny - self.mesh.overlap['top']) * self.mesh.nx)

    @property
    def _globalOverlappingCellIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2, 3] for mesh A

        ```
        ---------
        | 4 | 5 |
        ---------  B
        | 2 | 3 |
        =========
        | 0 | 1 |  A
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.offset[1] * self.mesh.nx, (self.mesh.offset[1] + self.mesh.ny) * self.mesh.nx)

    @property
    def _localNonOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of boundary cells.

        E.g., would return [0, 1] for mesh A

        ```
        ---------
        | 4 | 5 |
        ---------  B
        | 2 | 3 |
        =========
        | 0 | 1 |  A
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.overlap['bottom'] * self.mesh.nx,
                              (self.mesh.ny - self.mesh.overlap['top']) * self.mesh.nx)

    @property
    def _localOverlappingCellIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of boundary cells.

        E.g., would return [0, 1, 2, 3] for mesh A

        ```
        ---------
        |   |   |
        ---------  B
        | 2 | 3 |
        =========
        | 0 | 1 |  A
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(0, self.mesh.ny * self.mesh.nx)

    def _calcFaceIDs(self, y0, ny, global_horz_faces):
        prev_horz_faces = y0 * self.mesh.nx
        horz = numerix.arange(prev_horz_faces,
                              prev_horz_faces + (ny + 1) * self.mesh.nx)
        prev_vert_faces = y0 * (self.mesh.nx + 1)
        vert = numerix.arange(global_horz_faces + prev_vert_faces,
                              global_horz_faces + prev_vert_faces
                              + ny * (self.mesh.nx + 1))

        return numerix.concatenate((horz, vert))

    @property
    def _globalNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 8, 9, 10] for mesh A

        ```
        --6---7--
        14  15  16
        --4---5--  B
        11  12  13
        ==2===3==
        8   9   10 A
        --0---1--
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(y0=(self.mesh.offset[1]
                                     + self.mesh.overlap['bottom']),
                                 ny=(self.mesh.ny
                                     - self.mesh.overlap['bottom']
                                     - self.mesh.overlap['top']),
                                 global_horz_faces=(self.mesh.nx
                                                    * (self.mesh.args['ny'] + 1)))

    @property
    def _globalOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 13] for mesh A

        ```
        --6---7--
        14  15  16
        --4---5--  B
        11  12  13
        ==2===3==
        8   9   10 A
        --0---1--
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(y0=self.mesh.offset[1],
                                 ny=self.mesh.ny,
                                 global_horz_faces=(self.mesh.nx
                                                    * (self.mesh.args['ny'] + 1)))

    @property
    def _localNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 6, 7, 8] for mesh A
        and [2, 3, 4, 5, 6, 7, 11, 12, 13, 14, 15, 16] for mesh B

        ```
        --6---7--
        14  15  16
        --4---5--  B
        11  12  13
        ==2===3==
        6   7   8  A
        --0---1--
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(y0=self.mesh.overlap['bottom'],
                                 ny=(self.mesh.ny
                                     - self.mesh.overlap['bottom']
                                     - self.mesh.overlap['top']),
                                 global_horz_faces=0)

    @property
    def _localOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11] for mesh A
        and [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16] for mesh B

        ```
        --6---7--
        14  15  16
        --4---5--  B
        11  12  13
        ==2===3==
        6   7   8  A
        --0---1--
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(y0=0, ny=self.mesh.ny, global_horz_faces=0)

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell of grid"""
        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)
        cellTopology[:] = self._elementTopology["pixel"]

        return cellTopology


class _Grid3DTopology(_GridTopology):

    _concatenatedClass = Mesh

    @property
    def _globalNonOverlappingCellIDs(self):
        """
        Return the IDs of the local mesh in the context of the
        global parallel mesh. Does not include the IDs of boundary cells.

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange((self.mesh.offset[2] + self.mesh.overlap['front']) * self.mesh.nx * self.mesh.ny,
                              (self.mesh.offset[2] + self.mesh.nz - self.mesh.overlap['back']) * self.mesh.nx * self.mesh.ny)

    @property
    def _globalOverlappingCellIDs(self):
        """
        Return the IDs of the local mesh in the context of the
        global parallel mesh. Includes the IDs of boundary cells.

        .. note:: Trivial except for parallel meshes
        """

        return numerix.arange(self.mesh.offset[2] * self.mesh.nx * self.mesh.ny,
                              (self.mesh.offset[2] + self.mesh.nz) * self.mesh.nx * self.mesh.ny)

    @property
    def _localNonOverlappingCellIDs(self):
        """
        Return the IDs of the local mesh in isolation.
        Does not include the IDs of boundary cells.

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(self.mesh.overlap['front'] * self.mesh.nx * self.mesh.ny,
                              (self.mesh.nz - self.mesh.overlap['back']) * self.mesh.nx * self.mesh.ny)

    @property
    def _localOverlappingCellIDs(self):
        """
        Return the IDs of the local mesh in isolation.
        Includes the IDs of boundary cells.

        .. note:: Trivial except for parallel meshes
        """
        return numerix.arange(0, self.mesh.ny * self.mesh.nx * self.mesh.nz)

    def _calcFaceIDs(self, z0, nz, global_xy_faces, global_xz_faces):
        prev_xy_faces = z0 * self.mesh.nx * self.mesh.ny
        xy = numerix.arange(prev_xy_faces,
                            prev_xy_faces +
                            (nz + 1) * self.mesh.nx * self.mesh.ny)

        prev_xz_faces = z0 * self.mesh.nx * (self.mesh.ny + 1)
        xz = numerix.arange(global_xy_faces + prev_xz_faces,
                            global_xy_faces + prev_xz_faces +
                            self.mesh.nx * (self.mesh.ny + 1) * nz)

        prev_yz_faces = z0 * (self.mesh.nx + 1) * self.mesh.ny
        yz = numerix.arange(global_xy_faces + global_xz_faces + prev_yz_faces,
                            global_xy_faces + global_xz_faces + prev_yz_faces +
                            (self.mesh.nx + 1) * self.mesh.ny * nz)

        return numerix.concatenate((xy, xz, yz))

    @property
    def _globalNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 13, 14, 15
                            20, 21, 22, 23, 24, 25] for mesh A
        and [4, 5, 6, 7, 16, 17, 18, 19, 26, 27, 28] for mesh B

        ```
              ---------
             /   /   / |
            ---------| /
           /   /   / |/
          =========: / A
         /   /   / :/
        ---------  /         y
        |   |   | /  B       |
        ---------            --x
                            /
                           z
        XY    ---------
              | 0 | 1 |
            --------- -
            | 2 | 3 |
          ========= -  A
          | 4 | 5 |
        --------- =
        | 6 | 7 |    B
        ---------

        XZ    ---------
             / 10/ 11/
            --------- -
           / 14/ 15/ /
          ========= -  A
         / 18/ 19/ /
        --------- =
         / 16/ 17/   B
        ---------

        YZ   /|  /|  /|
            /2| /2| /2|
           /|0//|1//|2/
          /2|//2|//2|/ A
         /:3//:4//:5/
        /2://2://2:/
        |6/ |7/ |8/  B
        |/  |/  |/

              ---------
             / 10/ 11/2|
            ---------|2/
           / 14/ 15/2|/
          =========|5/ A
         / 18/ 19/:|/
        ---------2:/
        | 6 | 7 |8/  B
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(z0=(self.mesh.offset[2]
                                     + self.mesh.overlap['front']),
                                 nz=(self.mesh.nz
                                     - self.mesh.overlap['front']
                                     - self.mesh.overlap['back']),
                                 global_xy_faces=(self.mesh.nx
                                                  * self.mesh.ny
                                                  * (self.mesh.args['nz'] + 1)),
                                 global_xz_faces=(self.mesh.nx
                                                  * (self.mesh.ny + 1)
                                                  * self.mesh.args['nz']))

    @property
    def _globalOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in the context of the global parallel mesh.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 6, 7,
                            8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                            20, 21, 22, 23, 24, 25, 26, 27, 28] for mesh A
        and [2, 3, 4, 5, 6, 7, 12, 13, 14, 15, 16, 17, 18, 19,
             23, 24, 25, 26, 27, 28] for mesh B

        ```
              ---------
             /   /   / |
            ---------| /
           /   /   / |/
          =========: / A
         /   /   / :/
        ---------  /         y
        |   |   | /  B       |
        ---------            --x
                            /
                           z
        XY    ---------
              | 0 | 1 |
            --------- -
            | 2 | 3 |
          ========= -  A
          | 4 | 5 |
        --------- =
        | 6 | 7 |    B
        ---------

        XZ    ---------
             / 10/ 11/
            --------- -
           / 14/ 15/ /
          ========= -  A
         / 18/ 19/ /
        --------- =
         / 16/ 17/   B
        ---------

        YZ   /|  /|  /|
            /2| /2| /2|
           /|0//|1//|2/
          /2|//2|//2|/ A
         /:3//:4//:5/
        /2://2://2:/
        |6/ |7/ |8/  B
        |/  |/  |/

              ---------
             / 10/ 11/2|
            ---------|2/
           / 14/ 15/2|/
          =========|5/ A
         / 18/ 19/:|/
        ---------2:/
        | 6 | 7 |8/  B
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(z0=self.mesh.offset[2],
                                 nz=self.mesh.nz,
                                 global_xy_faces=(self.mesh.nx
                                                  * self.mesh.ny
                                                  * (self.mesh.args['nz'] + 1)),
                                 global_xz_faces=(self.mesh.nx
                                                  * (self.mesh.ny + 1)
                                                  * self.mesh.args['nz']))

    @property
    def _localNonOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Does not include the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 13, 14, 15
                            20, 21, 22, 23, 24, 25] for mesh A
        and [2, 3, 4, 5, 10, 11, 12, 13, 17, 18, 19] for mesh B

        ```
              ---------
             /   /   / |
            ---------| /
           /   /   / |/
          =========: / A
         /   /   / :/
        ---------  /         y
        |   |   | /  B       |
        ---------            --x
                            /
                           z
        XY    ---------
              | 0 | 1 |
            --------- -
            | 2 | 3 |
          ========= -  A
          |4/2|5/3|
        --------- =
        | 4 | 5 |    B
        ---------

        XZ    ---------
             / 10/ 11/
            --------- -
           / 14/ 15/ /
          ========= -  A
         / 12/ 13/ /
        --------- =
         / 10/ 11/   B
        ---------

        YZ   /|  /|  /|
            /2| /2| /2|
           /|0//|1//|2/
          /2|//2|//2|/ A
         /:3//:4//:5/
        /1://1://1:/
        |7/ |8/ |9/  B
        |/  |/  |/

              ---------
             / 10/ 11/2|
            ---------|2/
           / 14/ 15/2|/
          =========|5/ A
         / 12/ 13/:|/
        ---------1:/
        | 4 | 5 |9/  B
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(z0=self.mesh.overlap['front'],
                                 nz=(self.mesh.nz
                                     - self.mesh.overlap['front']
                                     - self.mesh.overlap['back']),
                                 global_xy_faces=0,
                                 global_xz_faces=0)

    @property
    def _localOverlappingFaceIDs(self):
        """Return the IDs of the local mesh in isolation.

        Includes the IDs of faces of boundary cells.

        E.g., would return [0, 1, 2, 3, 4, 5, 6, 7,
                            8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                            20, 21, 22, 23, 24, 25, 26, 27, 28] for mesh A
        and [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19] for mesh B

        ```
              ---------
             /   /   / |
            ---------| /
           /   /   / |/
          =========: / A
         /   /   / :/
        ---------  /         y
        |   |   | /  B       |
        ---------            --x
                            /
                           z
        XY    ---------
              | 0 | 1 |
            --------- -
            | 2 | 3 |
          ========= -  A
          |4/2|5/3|
        --------- =
        | 4 | 5 |    B
        ---------

        XZ    ---------
             / 10/ 11/
            --------- -
           / 14/ 15/ /
          ========= -  A
         / 12/ 13/ /
        --------- =
         / 10/ 11/   B
        ---------

        YZ   /|  /|  /|
            /2| /2| /2|
           /|0//|1//|2/
          /2|//2|//2|/ A
         /:3//:4//:5/
        /1://1://1:/
        |7/ |8/ |9/  B
        |/  |/  |/

              ---------
             / 10/ 11/2|
            ---------|2/
           / 14/ 15/2|/
          =========|5/ A
         / 12/ 13/:|/
        ---------1:/
        | 4 | 5 |9/  B
        ---------
        ```

        .. note:: Trivial except for parallel meshes
        """
        return self._calcFaceIDs(z0=0,
                                 nz=self.mesh.nz,
                                 global_xy_faces=0,
                                 global_xz_faces=0)

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell of grid"""
        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)
        cellTopology[:] = self._elementTopology["voxel"]

        return cellTopology

class _PeriodicGrid1DTopology(_Grid1DTopology):

    @property
    def _globalOverlappingCellIDs(self):
        return super(_PeriodicGrid1DTopology, self)._globalOverlappingCellIDs % self.mesh.args['nx']



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix
from fipy.terms.unaryTerm import _UnaryTerm
from fipy.terms import TermMultiplyError

class _NonDiffusionTerm(_UnaryTerm):

    def __neg__(self):
        r"""
         Negate a `Term`.

           >>> -__NonDiffusionTerm(coeff=1.)
           __NonDiffusionTerm(coeff=-1.0)

        """
        if isinstance(self.coeff, (tuple, list)):
            return self.__class__(coeff=-numerix.array(self.coeff), var=self.var)
        else:
            return self.__class__(coeff=-self.coeff, var=self.var)

    def __mul__(self, other):
        r"""
        Multiply a term

            >>> 2. * __NonDiffusionTerm(coeff=0.5)
            __NonDiffusionTerm(coeff=1.0)

        Test for ticket:291.

            >>> from fipy import PowerLawConvectionTerm
            >>> PowerLawConvectionTerm(coeff=[[1], [0]]) * 1.0
            PowerLawConvectionTerm(coeff=array([[ 1.],
                   [ 0.]]))

        """

        if isinstance(other, (int, float)):
            if isinstance(self.coeff, (list, tuple)):
                coeff = numerix.array(self.coeff)
            else:
                coeff = self.coeff
            return self.__class__(coeff=other * coeff, var=self.var)
        else:
            raise TermMultiplyError

    __rmul__ = __mul__

    @property
    def _diffusionVars(self):
        return []

    def _getDiagonalSign(self, transientGeomCoeff=None, diffusionGeomCoeff=None):
        if transientGeomCoeff is not None and diffusionGeomCoeff is not None:
            diagonalSign = numerix.where(numerix.array(numerix.all(transientGeomCoeff == 0, axis=-1)),
                                         numerix.array(2 * numerix.all(diffusionGeomCoeff[0] <= 0, axis=-1) - 1),
                                         numerix.array(2 * numerix.all(transientGeomCoeff >= 0, axis=-1) - 1))
        elif transientGeomCoeff is not None:
            diagonalSign = 2 * numerix.all(transientGeomCoeff >= 0, axis=-1) - 1
        elif diffusionGeomCoeff is not None:
            diagonalSign = 2 * numerix.all(diffusionGeomCoeff[0] <= 0, axis=-1) - 1
        else:
            diagonalSign = 1

        return diagonalSign

    def _test(self):
        r"""
        Test stuff.

         Subtract a `Term` from a `Term`, number or variable.

           >>> __NonDiffusionTerm(coeff=1.) - 10.
           (__NonDiffusionTerm(coeff=1.0) + -10.0)
           >>> __NonDiffusionTerm(coeff=1.) - __NonDiffusionTerm(coeff=2.)
           (__NonDiffusionTerm(coeff=1.0) + __NonDiffusionTerm(coeff=-2.0))

         Subtract a `Term`, number or variable from a `Term`.

           >>> 10. - __NonDiffusionTerm(coeff=1.)
           (__NonDiffusionTerm(coeff=-1.0) + 10.0)

        Add a `Term` to another `Term`, number or variable.

           >>> __NonDiffusionTerm(coeff=1.) + 10.
           (__NonDiffusionTerm(coeff=1.0) + 10.0)
           >>> __NonDiffusionTerm(coeff=1.) + __NonDiffusionTerm(coeff=2.)
           (__NonDiffusionTerm(coeff=1.0) + __NonDiffusionTerm(coeff=2.0))
           >>> 10. + __NonDiffusionTerm(coeff=1.)
           (__NonDiffusionTerm(coeff=1.0) + 10.0)

        Positive of a `Term`.

           >>> +__NonDiffusionTerm(coeff=1.)
           __NonDiffusionTerm(coeff=1.0)

        This method allows `Terms` to be equated in a natural way. Note that the
        following does not return `False.`

           >>> __NonDiffusionTerm(coeff=1.) == __NonDiffusionTerm(coeff=2.)
           (__NonDiffusionTerm(coeff=1.0) + __NonDiffusionTerm(coeff=-2.0))

        it is equivalent to,

           >>> __NonDiffusionTerm(coeff=1.) - __NonDiffusionTerm(coeff=2.)
           (__NonDiffusionTerm(coeff=1.0) + __NonDiffusionTerm(coeff=-2.0))

        A `Term` can also equate with a number.

           >>> __NonDiffusionTerm(coeff=1.) == 1.
           (__NonDiffusionTerm(coeff=1.0) + -1.0)

        Likewise for integers.

           >>> __NonDiffusionTerm(coeff=1.) == 1
           (__NonDiffusionTerm(coeff=1.0) + -1)

        Equating to zero is allowed, of course

            >>> __NonDiffusionTerm(coeff=1.) == 0
            __NonDiffusionTerm(coeff=1.0)
            >>> 0 == __NonDiffusionTerm(coeff=1.)
            __NonDiffusionTerm(coeff=1.0)

        Divide a term

            >>> __NonDiffusionTerm(2.) / 2.
            __NonDiffusionTerm(coeff=1.0)

        Combine this equation with another

            >>> from fipy.variables.variable import Variable
            >>> eq1 = 10. + __NonDiffusionTerm(coeff=1., var=Variable(name='A'))
            >>> eq2 = 20. + __NonDiffusionTerm(coeff=2., var=Variable(name='B'))
            >>> eq1 & eq2
            ((__NonDiffusionTerm(coeff=1.0, var=A) + 10.0) & (__NonDiffusionTerm(coeff=2.0, var=B) + 20.0))

        """



class __NonDiffusionTerm(_NonDiffusionTerm):
    """
    Dummy subclass for tests
    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = ["Stepper"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Stepper(object):
    def __init__(self, vardata=()):
        self.vardata = vardata

    def sweepFn(vardata, dt, *args, **kwargs):
        residual = 0
        for var, eqn, bcs in vardata:
            residual = max(residual, eqn.sweep(var=var, dt=dt, boundaryConditions=bcs))

        return residual
    sweepFn = staticmethod(sweepFn)

    def successFn(vardata, dt, dtPrev, elapsed, *args, **kwargs):
        pass
    successFn = staticmethod(successFn)

    def failFn(vardata, dt, *args, **kwargs):
        pass
    failFn = staticmethod(failFn)

    def _lowerBound(self, dt):
        dt = max(dt, self.dtMin)
        if self.elapsed + dt == self.elapsed:
            raise FloatingPointError("step size underflow: %g + %g == %g" % (self.elapsed, dt, self.elapsed))

        return dt

    def _step(self, dt, dtPrev, sweepFn, failFn, *args, **kwargs):
        sweepFn(vardata=self.vardata, dt=dt, *args, **kwargs)
        return dt, dt

    def step(self, dt, dtTry=None, dtMin=None, dtPrev=None,
             sweepFn=None, successFn=None, failFn=None, *args, **kwargs):
        sweepFn = sweepFn or self.sweepFn
        successFn = successFn or self.successFn
        failFn = failFn or self.failFn

        dtTry = dtTry or dtMin or dt
        dtPrev = dtPrev or dtMin
        self.dtMin = dtMin or 0.

        self.elapsed = 0.

        while self.elapsed < dt:
            dtMax = dt - self.elapsed
            if dtTry > dtMax:
                dtSave = dtTry
                dtTry = dtMax
            else:
                dtSave = None

            for var, eqn, bcs in self.vardata:
                var.updateOld()

            dtPrev, dtTry = self._step(dt=dtTry, dtPrev=dtPrev,
                                       sweepFn=sweepFn, failFn=failFn,
                                       *args, **kwargs)

            self.elapsed += dtPrev

            successFn(vardata=self.vardata,
                      dtPrev=dtPrev, elapsed=self.elapsed, dt=dt, *args, **kwargs)

            dtTry = max(dtTry, self.dtMin)

        return dtSave or dtPrev, dtSave or dtTry



"""
Generic mesh class using numerix to do the calculations

Meshes contain cells, faces, and vertices.

This is built for a non-mixed element mesh.
"""
from __future__ import division
from __future__ import unicode_literals
from builtins import range
from builtins import zip
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools import serialComm

from fipy.meshes.mesh import Mesh
from fipy.meshes.representations.meshRepresentation import _MeshRepresentation
from fipy.meshes.topologies.meshTopology import _Mesh2DTopology

def _orderVertices(vertexCoords, vertices):
    coordinates = numerix.take(vertexCoords, vertices)
    centroid = numerix.add.reduce(coordinates) / coordinates.shape[0]
    coordinates = coordinates - centroid
    # to prevent division by zero
    coordinates = numerix.where(coordinates == 0, 1.e-100, coordinates)
    # angles go from -pi / 2 to 3*pi / 2
    angles = numerix.arctan(coordinates[:, 1] / coordinates[:, 0]) \
               + numerix.where(coordinates[:, 0] < 0, numerix.pi, 0)
    sortorder = numerix.argsort(angles)
    return numerix.take(vertices, sortorder)

__all__ = ["Mesh2D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class Mesh2D(Mesh):
    def __init__(self, vertexCoords, faceVertexIDs, cellFaceIDs, communicator=serialComm, _RepresentationClass=_MeshRepresentation, _TopologyClass=_Mesh2DTopology):
        super(Mesh2D, self).__init__(vertexCoords=vertexCoords, faceVertexIDs=faceVertexIDs, cellFaceIDs=cellFaceIDs, communicator=communicator,
                                     _RepresentationClass=_RepresentationClass, _TopologyClass=_TopologyClass)

    def _calcScaleArea(self):
        return self.scale['length']

    def _calcScaleVolume(self):
        return self.scale['length']**2

    def _calcFaceAreas(self):
        faceVertexCoords = numerix.take(self.vertexCoords, self.faceVertexIDs, axis=1)
        tangent = faceVertexCoords[:, 1] - faceVertexCoords[:, 0]
        return numerix.sqrtDot(tangent, tangent)

    def _calcFaceNormals(self):
        faceVertexCoords = numerix.take(self.vertexCoords, self.faceVertexIDs, axis=1)
        t1 = faceVertexCoords[:, 1,:] - faceVertexCoords[:, 0,:]
        faceNormals = t1.copy()
        mag = numerix.sqrt(t1[1]**2 + t1[0]**2)
        faceNormals[0] = -t1[1] / mag
        faceNormals[1] = t1[0] / mag

        orientation = 1 - 2 * (numerix.dot(faceNormals, self.cellDistanceVectors) < 0)
        return faceNormals * orientation

    def _calcFaceTangents(self):
        # copy required to get internal memory ordering correct for inlining.
        faceTangents1 = numerix.array((-self.faceNormals[1], self.faceNormals[0])).copy()
        faceTangents2 = numerix.zeros(faceTangents1.shape, 'd')
        return faceTangents1, faceTangents2

    def _translate(self, vector):
        newCoords = self.vertexCoords + vector
        newmesh = Mesh2D(newCoords, self.faceVertexIDs, self.cellFaceIDs)
        return newmesh

    def __mul__(self, factor):
        newCoords = self.vertexCoords * factor
        newmesh = Mesh2D(newCoords, self.faceVertexIDs, self.cellFaceIDs)
        return newmesh

    def _calcOrderedCellVertexIDs(self):
        from fipy.tools.numerix import take
        NFac = self._maxFacesPerCell

        # numpy 1.1's MA.take doesn't like FlatIter. Call ravel() instead.
        cellVertexIDs0 = take(self.faceVertexIDs[0], self.cellFaceIDs.ravel())
        cellVertexIDs1 = take(self.faceVertexIDs[1], self.cellFaceIDs.ravel())
        cellVertexIDs = MA.where(self._cellToFaceOrientations.ravel() > 0,
                             cellVertexIDs0, cellVertexIDs1)

        cellVertexIDs = numerix.reshape(cellVertexIDs, (NFac, -1))
        return cellVertexIDs

    @property
    def _nonOrthogonality(self):

        exteriorFaceArray = numerix.zeros((self.faceCellIDs.shape[1],), 'l')
        numerix.put(exteriorFaceArray, numerix.nonzero(self.exteriorFaces), 1)
        unmaskedFaceCellIDs = MA.filled(self.faceCellIDs, 0)
        # what we put in for the "fill" doesn't matter because only exterior
        # faces have anything masked, and exterior faces have their displacement
        # vectors set to zero.
        #
        # if it's an exterior face, make the "displacement vector" equal to zero
        # so the cross product will be zero.

        faceDisplacementVectors = \
          numerix.where(numerix.array(list(zip(exteriorFaceArray, exteriorFaceArray))),
                        0.0,
                        numerix.take(self._scaledCellCenters.swapaxes(0, 1),
                                     unmaskedFaceCellIDs[1,:]) \
                          - numerix.take(self._scaledCellCenters.swapaxes(0, 1),
                        unmaskedFaceCellIDs[0,:]))

        faceDisplacementVectors = faceDisplacementVectors.swapaxes(0, 1)

        faceCrossProducts = (faceDisplacementVectors[0,:] * self.faceNormals[1,:]) \
          - (faceDisplacementVectors[1,:] * self.faceNormals[0,:])

        faceDisplacementVectorLengths = numerix.maximum(((faceDisplacementVectors[0,:] ** 2) \
          + (faceDisplacementVectors[1,:] ** 2)) ** 0.5, 1.e-100)

        faceWeightedNonOrthogonalities = abs(faceCrossProducts / faceDisplacementVectorLengths) * self._faceAreas

        cellFaceWeightedNonOrthogonalities = numerix.take(faceWeightedNonOrthogonalities, self.cellFaceIDs)

        cellFaceAreas = numerix.take(self._faceAreas, self.cellFaceIDs)
        cellTotalWeightedValues = numerix.add.reduce(cellFaceWeightedNonOrthogonalities, axis = 0)
        cellTotalFaceAreas = numerix.add.reduce(cellFaceAreas, axis = 0)

        return (cellTotalWeightedValues / cellTotalFaceAreas)

    def extrude(self, extrudeFunc=lambda x: x + numerix.array((0, 0, 1))[:, numerix.newaxis] , layers=1):
        """
        This function returns a new 3D mesh. The 2D mesh is extruded
        using the `extrudeFunc` and the number of layers.

        >>> from fipy.meshes.nonUniformGrid2D import NonUniformGrid2D
        >>> print(NonUniformGrid2D(nx=2, ny=2).extrude(layers=2).cellCenters)
        [[ 0.5  1.5  0.5  1.5  0.5  1.5  0.5  1.5]
         [ 0.5  0.5  1.5  1.5  0.5  0.5  1.5  1.5]
         [ 0.5  0.5  0.5  0.5  1.5  1.5  1.5  1.5]]

        >>> from fipy.meshes.tri2D import Tri2D
        >>> print(Tri2D().extrude(layers=2).cellCenters.allclose([[ 0.83333333, 0.5,        0.16666667, 0.5,       0.83333333, 0.5,
        ...                                                      0.16666667, 0.5       ],
        ...                                                       [ 0.5,        0.83333333, 0.5,        0.16666667, 0.5,        0.83333333,
        ...                                                      0.5,        0.16666667],
        ...                                                      [ 0.5,        0.5,        0.5,        0.5,        1.5,        1.5,        1.5,
        ...                                                      1.5       ]]))
        True

        Parameters
        ----------
        extrudeFunc : function
            Takes the vertex coordinates and returns the displaced values
        layers : int
            Number of layers in the extruded mesh (number of times
            `extrudeFunc` will be called)
        """

        return self._extrude(self, extrudeFunc, layers)

    def _extrude(self, mesh, extrudeFunc, layers):
        ## should extrude self rather than creating a new mesh?

        ## the following allows the 2D mesh to be in 3D space, this can be the case for a
        ## Gmsh2DIn3DSpace which would then be extruded.
        oldVertices = mesh.vertexCoords
        if oldVertices.shape[0] == 2:
            oldVertices = numerix.resize(oldVertices, (3, len(oldVertices[0])))
            oldVertices[2] = 0

        NCells = mesh.numberOfCells
        NFac = mesh.numberOfFaces
        NFacPerCell =  mesh._maxFacesPerCell

        ## set up the initial data arrays
        new_shape = (max(NFacPerCell, 4), (1 + layers)*NCells + layers*NFac)
        faces = numerix.MA.masked_values(-numerix.ones(new_shape, 'l'), value = -1)
        orderedVertices = mesh._orderedCellVertexIDs
        faces[:NFacPerCell, :NCells] = orderedVertices
        vertices = oldVertices
        vert0 = mesh.faceVertexIDs
        faceCount = NCells

        for layer in range(layers):

            ## need this later
            initialFaceCount = faceCount

            ## build the vertices
            newVertices = extrudeFunc(oldVertices)
            vertices = numerix.concatenate((vertices, newVertices), axis=1)

            ## build the faces along the layers
            faces[:NFacPerCell, faceCount: faceCount + NCells] = orderedVertices + len(oldVertices[0]) * (layer + 1)
            try:
                # numpy 1.1 doesn't copy right side before assigning slice
                # See: http://www.mail-archive.com/numpy-discussion@scipy.org/msg09843.html
                faces[:NFacPerCell, faceCount: faceCount + NCells] = faces[:NFacPerCell, faceCount: faceCount + NCells][::-1,:].copy()
            except:
                faces[:NFacPerCell, faceCount: faceCount + NCells] = faces[:NFacPerCell, faceCount: faceCount + NCells][::-1,:]

            faceCount = faceCount + NCells

            vert1 = (vert0 + len(oldVertices[0]))[::-1,:]

            ## build the faces between the layers
            faces[:4, faceCount: faceCount + NFac] = numerix.concatenate((vert0, vert1), axis = 0)[::-1,:]

            vert0 = vert0 + len(oldVertices[0])

            NCells = mesh.numberOfCells


            ## build the cells, the first layer has slightly different ordering
            if layer == 0:
                c0 =  numerix.reshape(numerix.arange(NCells), (1, NCells))
                cells = numerix.concatenate((c0, c0 + NCells, mesh.cellFaceIDs + 2 * NCells), axis = 0)
            else:
                newCells = numerix.concatenate((c0, c0 + initialFaceCount, mesh.cellFaceIDs + faceCount), axis=0)
                newCells[0] = cells[1, -NCells:]
                cells = numerix.concatenate((cells, newCells), axis=1)

            ## keep a count of things for the next layer
            faceCount = faceCount + NFac
            oldVertices = newVertices

        ## return a new mesh, extrude could just as easily act on self
        return Mesh(vertices, faces, cells, communicator=mesh.communicator)

    @property
    def _VTKCellType(self):
        try:
            from tvtk.api import tvtk
        except ImportError as e:
            from enthought.tvtk.api import tvtk
        return tvtk.Polygon().cell_type

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected.

            >>> dx = 0.5
            >>> dy = 2.
            >>> nx = 3
            >>> ny = 2

            >>> vertices = numerix.array(((0., 1., 2., 3.,
            ...                            0., 1., 2., 3.,
            ...                            0., 1., 2., 3., 4.),
            ...                           (0., 0., 0., 0.,
            ...                            1., 1., 1., 1.,
            ...                            2., 2., 2., 2., 1.)))
            >>> vertices *= numerix.array(((dx,), (dy,)))
            >>> faces = numerix.array(((1, 2, 3, 4, 5, 6, 8,  9, 10,
            ...                         0, 5, 6, 7, 4, 9, 10, 11, 12,  7, 11),
            ...                        (0, 1, 2, 5, 6, 7, 9, 10, 11,
            ...                         4, 1, 2, 3, 8, 5,  6,  7,  3, 12, 12)))
            >>> cells = MA.masked_values((( 0,  1,  2,  3,  4,  5, 17, 18),
            ...                           (10, 11, 12, 14, 15, 16, 18, 19),
            ...                           ( 3,  4,  5,  6,  7,  8, 12, 16),
            ...                           ( 9, 10, 11, 13, 14, 15, -1, -1)), -1)

            >>> mesh = Mesh2D(vertexCoords = vertices, faceVertexIDs = faces, cellFaceIDs = cells)

            >>> externalFaces = numerix.array((0, 1, 2, 6, 7, 8, 9, 13, 17, 19))
            >>> print(numerix.allequal(externalFaces,
            ...                        numerix.nonzero(mesh.exteriorFaces)))
            1

            >>> internalFaces = numerix.array((3, 4, 5, 10, 11, 12, 14, 15, 16, 18))
            >>> print(numerix.allequal(internalFaces,
            ...                        numerix.nonzero(mesh.interiorFaces)))
            1

            >>> faceCellIds = MA.masked_values((( 0,  1,  2, 0,  1, 2,  3,  4,  5,
            ...                                   0,  0,  1, 2,  3, 3,  4,  5,  6, 6,  7),
            ...                                 (-1, -1, -1, 3,  4, 5, -1, -1, -1,
            ...                                  -1,  1,  2, 6, -1, 4,  5,  7, -1, 7, -1)), -1)
            >>> numerix.allequal(faceCellIds, mesh.faceCellIDs)
            1

            >>> faceAreas = numerix.array((dx, dx, dx, dx, dx, dx, dx, dx, dx,
            ...                            dy, dy, dy, dy, dy, dy, dy, dy,
            ...                            numerix.sqrt(dx**2 + dy**2), dx, numerix.sqrt(dx**2 + dy**2)))
            >>> numerix.allclose(faceAreas, mesh._faceAreas, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceCoords = numerix.take(vertices, faces, axis=1)
            >>> faceCenters = (faceCoords[..., 0,:] + faceCoords[..., 1,:]) / 2.
            >>> print(numerix.allclose(faceCenters, mesh.faceCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> faceNormals = numerix.array(((0., 0., 0., 0., 0., 0., 0., 0., 0.,
            ...                               -1., 1., 1., 1., -1., 1., 1., 1.,
            ...                               dy / numerix.sqrt(dx**2 + dy**2),
            ...                               0.,
            ...                               dy / numerix.sqrt(dx**2 + dy**2)),
            ...                              (-1., -1., -1., 1., 1., 1., 1., 1., 1.,
            ...                               0., 0., 0., 0., 0., 0., .0, 0.,
            ...                               -dx / numerix.sqrt(dx**2 + dy**2),
            ...                               1.,
            ...                               dx / numerix.sqrt(dx**2 + dy**2))))
            >>> numerix.allclose(faceNormals, mesh.faceNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToFaceOrientations = MA.masked_values(((1,  1,  1, -1, -1, -1,  1, -1),
            ...                                            (1,  1,  1,  1,  1,  1,  1,  1),
            ...                                            (1,  1,  1,  1,  1,  1, -1, -1),
            ...                                            (1, -1, -1,  1, -1, -1,  0,  0)), 0)
            >>> numerix.allequal(cellToFaceOrientations, mesh._cellToFaceOrientations)
            1

            >>> cellVolumes = numerix.array((dx*dy, dx*dy, dx*dy, dx*dy, dx*dy, dx*dy, dx*dy / 2., dx*dy / 2.))
            >>> numerix.allclose(cellVolumes, mesh.cellVolumes, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellCenters = numerix.array(((dx/2., 3.*dx/2., 5.*dx/2., dx/2., 3.*dx/2., 5.*dx/2., 3.*dx+dx/3., 3.*dx+dx/3.),
            ...                              (dy/2., dy/2., dy/2., 3.*dy/2., 3.*dy/2., 3.*dy/2., 2.*dy/3., 4.*dy/3.)))
            >>> print(numerix.allclose(cellCenters, mesh.cellCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> faceToCellDistances = MA.masked_values(((dy / 2., dy / 2., dy / 2.,
            ...                                          dy / 2., dy / 2., dy / 2.,
            ...                                          dy / 2., dy / 2., dy / 2.,
            ...                                          dx / 2., dx / 2., dx / 2., dx / 2.,
            ...                                          dx / 2., dx / 2., dx / 2., dx / 2.,
            ...                                          numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2),
            ...                                          numerix.sqrt((dx / 6.)**2 + (dy / 3.)**2),
            ...                                          numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2)),
            ...                                         (-1, -1, -1,
            ...                                          dy / 2., dy / 2., dy / 2.,
            ...                                          -1, -1, -1,
            ...                                          -1, dx / 2., dx / 2., numerix.sqrt((dx / 3.)**2 + (dy / 6.)**2),
            ...                                          -1, dx / 2., dx / 2., numerix.sqrt((dx / 3.)**2 + (dy / 6.)**2),
            ...                                          -1,
            ...                                          numerix.sqrt((dx / 6.)**2 + (dy / 3.)**2),
            ...                                          -1)), -1)
            >>> numerix.allclose(faceToCellDistances, mesh._faceToCellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellDistances = numerix.array((dy / 2., dy / 2., dy / 2.,
            ...                                dy, dy, dy,
            ...                                dy / 2., dy / 2., dy / 2.,
            ...                                dx / 2., dx, dx,
            ...                                numerix.sqrt((dx / 3. + dx / 2.)**2 + (dy / 6.)**2),
            ...                                dx / 2., dx, dx,
            ...                                numerix.sqrt((dx / 3. + dx / 2.)**2 + (dy / 6.)**2),
            ...                                numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2),
            ...                                2. * dy / 3.,
            ...                                numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2)))
            >>> numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)
            1

            >>> areaProjections = faceNormals * faceAreas
            >>> numerix.allclose(areaProjections, mesh._areaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> tangents1 = numerix.array(((1., 1., 1., -1., -1., -1., -1., -1., -1.,
            ...                             0., 0., 0., 0., 0., 0., 0., 0.,
            ...                             dx / numerix.sqrt(dx**2 +dy**2),
            ...                             -1.,
            ...                             -dx / numerix.sqrt(dx**2 +dy**2)),
            ...                            (0., 0., 0., 0., 0., 0., 0., 0., 0.,
            ...                             -1., 1., 1., 1., -1., 1., 1., 1.,
            ...                             dy / numerix.sqrt(dx**2 +dy**2),
            ...                             0.,
            ...                             dy / numerix.sqrt(dx**2 +dy**2))))
            >>> numerix.allclose(tangents1, mesh._faceTangents1, atol = 1e-10, rtol = 1e-10)
            1

            >>> tangents2 = numerix.array(((0., 0., 0., 0., -0., -0., -0., -0., -0.,
            ...                             -0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.),
            ...                            (0., 0., 0., 0., 0., 0., 0., 0., 0.,
            ...                             -0., 0., 0., 0., -0., 0., 0., 0., 0., 0., 0.)))
            >>> numerix.allclose(tangents2, mesh._faceTangents2, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToCellIDs = MA.masked_values(((-1, -1, -1,  0,  1,  2, -1,  6),
            ...                                   ( 1,  2,  6,  4,  5,  7,  7, -1),
            ...                                   ( 3,  4,  5, -1, -1, -1,  2,  5),
            ...                                   (-1,  0,  1, -1,  3,  4, -1, -1)), -1)
            >>> numerix.allequal(cellToCellIDs, mesh._cellToCellIDs)
            1

            >>> d1 = numerix.sqrt((5. * dx / 6.)**2 + (dy / 6.)**2)
            >>> d2 = numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2)
            >>> cellToCellDistances = MA.masked_values(((dy / 2., dy / 2., dy / 2., dy, dy, dy, d2, 2. * dy / 3.),
            ...                                         (dx, dx, d1, dx, dx, d1, 2. * dy / 3., d2),
            ...                                         (dy, dy, dy, dy / 2., dy / 2., dy / 2., d1, d1),
            ...                                         (dx / 2., dx, dx, dx / 2., dx, dx, -1, -1)), -1)
            >>> numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> interiorCellIDs = numerix.array(())
            >>> numerix.allequal(interiorCellIDs, mesh._interiorCellIDs)
            1

            >>> exteriorCellIDs = numerix.array((0, 1, 2, 3, 4, 5, 6, 7))
            >>> numerix.allequal(exteriorCellIDs, mesh._exteriorCellIDs)
            1

            >>> nx = dy / numerix.sqrt(dx**2 + dy**2)
            >>> ny = dx / numerix.sqrt(dx**2 + dy**2)
            >>> cellNormals = MA.masked_values(((( 0,  0,  0,  0,  0,  0,    nx,     0),
            ...                                  ( 1,  1,  1,  1,  1,  1,     0,    nx),
            ...                                  ( 0,  0,  0,  0,  0,  0,    -1,    -1),
            ...                                  (-1, -1, -1, -1, -1, -1, -1000, -1000)),
            ...                                 ((-1, -1, -1, -1, -1, -1,   -ny,    -1),
            ...                                  ( 0,  0,  0,  0,  0,  0,     1,    ny),
            ...                                  ( 1,  1,  1,  1,  1,  1,     0,     0),
            ...                                  ( 0,  0,  0,  0,  0,  0, -1000, -1000))), -1000)
            >>> numerix.allclose(cellNormals, mesh._cellNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> area = numerix.sqrt(dx**2 + dy**2)
            >>> cellAreaProjections = MA.masked_values((((  0,  0,  0,  0,  0,  0,  nx * area,         0),
            ...                                          ( dy, dy, dy, dy, dy, dy,          0, nx * area),
            ...                                          (  0,  0,  0,  0,  0,  0,        -dy,       -dy),
            ...                                          (-dy, -dy, -dy, -dy, -dy, -dy,      -1000,     -1000)),
            ...                                         ((-dx, -dx, -dx, -dx, -dx, -dx, -ny * area,       -dx),
            ...                                          (  0,  0,  0,  0,  0,  0,         dx, ny * area),
            ...                                          ( dx, dx, dx, dx, dx, dx,          0,         0),
            ...                                          (  0,  0,  0,  0,  0,  0,      -1000,     -1000))), -1000)
            >>> numerix.allclose(cellAreaProjections, mesh._cellAreaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellVertexIDs = MA.masked_array(((5, 6, 7, 9, 10, 11,    12,    12),
            ...                                  (4, 5, 6, 8,  9, 10,     7,    11),
            ...                                  (1, 2, 3, 5,  6,  7,     3,     7),
            ...                                  (0, 1, 2, 4,  5,  6, -1000, -1000)), -1000)

            >>> numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)
            1


            >>> from fipy.tools import dump
            >>> (f, filename) = dump.write(mesh, extension = '.gz')
            >>> unpickledMesh = dump.read(filename, f)

            >>> print(numerix.allequal(mesh.cellCenters, unpickledMesh.cellCenters))
            True



        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
__all__ = []

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    theSuite = _LateImportDocTestSuite(docTestModuleNames = (
            'dimensions.physicalField',
            'numerix',
            'dump',
            'vector',
            'sharedtempfile'
        ), base = __name__)

    return theSuite

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix

from fipy.meshes.topologies.abstractTopology import _AbstractTopology

class _MeshTopology(_AbstractTopology):

    @property
    def _concatenatedClass(self):
        from fipy.meshes.mesh import Mesh
        return Mesh

    @property
    def _isOrthogonal(self):
        return False

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell"""
        facesPerCell = self.mesh._facesPerCell
        nodesPerFace = self.mesh._nodesPerFace

        def faceCountsMatch(targetCounts):
            if len(targetCounts) > nodesPerFace.shape[0]:
                # pad nodesPerFace with zeros
                paddedNodesPerFace = numerix.zeros((len(targetCounts), nodesPerFace.shape[1]), dtype=numerix.INT_DTYPE)
                paddedNodesPerFace[:nodesPerFace.shape[0],:] = nodesPerFace

                paddedTargetCounts = numerix.array(targetCounts)[..., numerix.newaxis]
            else:
                # pad target face node count with zeros
                paddedTargetCounts = numerix.concatenate((targetCounts,
                                                          [0] * (self.mesh._maxFacesPerCell - len(targetCounts))))
                paddedTargetCounts = paddedTargetCounts[..., numerix.newaxis]

                paddedNodesPerFace = nodesPerFace

            return ((facesPerCell == len(targetCounts))
                    & (paddedNodesPerFace == paddedTargetCounts).all(axis=0))

        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)

        t = self._elementTopology

        if self.mesh.dim == 1:
            cellTopology[:] = t["line"]
        elif self.mesh.dim == 2:
            cellTopology[:] = t["polygon"]
            cellTopology[faceCountsMatch([2, 2, 2])] = t["triangle"]
            cellTopology[faceCountsMatch([2, 2, 2, 2])] = t["quadrangle"]
        else:
            cellTopology[:] = t["unknown"]
            cellTopology[faceCountsMatch([3, 3, 3, 3])] = t["tetrahedron"]
            cellTopology[faceCountsMatch([4, 4, 4, 4, 4, 4])] = t["hexahedron"]
            cellTopology[faceCountsMatch([4, 4, 4, 3, 3])] = t["prism"]
            cellTopology[faceCountsMatch([4, 3, 3, 3, 3])] = t["pyramid"]

        return cellTopology

class _Mesh1DTopology(_MeshTopology):

    @property
    def _concatenatedClass(self):
        from fipy.meshes.mesh1D import Mesh1D
        return Mesh1D

    @property
    def _isOrthogonal(self):
        return True

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell of grid"""
        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)
        cellTopology[:] = self._elementTopology["line"]

        return cellTopology

class _Mesh2DTopology(_MeshTopology):

    @property
    def _concatenatedClass(self):
        from fipy.meshes.mesh2D import Mesh2D
        return Mesh2D

    @property
    def _cellTopology(self):
        """return a map of the topology of each cell"""
        cellTopology = numerix.empty((self.mesh.numberOfCells,), dtype=numerix.ubyte)

        t = self._elementTopology
        cellTopology[:] = t["polygon"]

        facesPerCell = self.mesh._facesPerCell
        cellTopology[facesPerCell == 3] = t["triangle"]
        cellTopology[facesPerCell == 4] = t["quadrangle"]

        return cellTopology



from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["ExponentialNoiseVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class ExponentialNoiseVariable(NoiseVariable):
    r"""
    Represents an exponential distribution of random numbers with the probability
    distribution

    .. math::

       \mu^{-1} e^{-\frac{x}{\mu}}

    with a mean parameter :math:`\mu`.

    Seed the random module for the sake of deterministic test results.

    >>> from fipy import numerix
    >>> numerix.random.seed(1)

    We generate noise on a uniform Cartesian mesh

    >>> from fipy.variables.variable import Variable
    >>> mean = Variable()
    >>> from fipy.meshes import Grid2D
    >>> noise = ExponentialNoiseVariable(mesh = Grid2D(nx = 100, ny = 100), mean = mean)

    We histogram the root-volume-weighted noise distribution

    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution = noise, dx = 0.1, nx = 100)

    and compare to a Gaussian distribution

    >>> from fipy.variables.cellVariable import CellVariable
    >>> expdist = CellVariable(mesh = histogram.mesh)
    >>> x = histogram.mesh.cellCenters[0]

    >>> if __name__ == '__main__':
    ...     from fipy import Viewer
    ...     viewer = Viewer(vars=noise, datamin=0, datamax=5)
    ...     histoplot = Viewer(vars=(histogram, expdist),
    ...                        datamin=0, datamax=1.5)

    >>> from fipy.tools.numerix import arange, exp

    >>> for mu in arange(0.5, 3, 0.5):
    ...     mean.value = (mu)
    ...     expdist.value = ((1/mean)*exp(-x/mean))
    ...     if __name__ == '__main__':
    ...         import sys
    ...         print("mean: %g" % mean, file=sys.stderr)
    ...         viewer.plot()
    ...         histoplot.plot()

    >>> print(abs(noise.faceGrad.divergence.cellVolumeAverage) < 5e-15)
    1

    .. image:: fipy/variables/exp.*
      :scale: 25
      :align: center
      :alt: random values with an exponential distribution

    .. image:: fipy/variables/exp-histogram.*
      :scale: 25
      :align: center
      :alt: histogram of random values with an exponential distribution

    """
    def __init__(self, mesh, mean=0.0, name = '', hasOld = 0):
        r"""
        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The mesh on which to define the noise.
        mean : float
            The mean of the distribution :math:`\mu`.
        """
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)
        self.mean = self._requires(mean)

    def random(self):
        return random.exponential(scale = self.mean,
                                  size = [self.mesh.globalNumberOfCells])

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearBicgSolver"]

class LinearBicgSolver(PETScKrylovSolver):

    """
    The `LinearBicgSolver` is an interface to the biconjugate gradient solver in
    PETSc, using no preconditioner by default.
    """
      
    solver = 'bicg'



from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from builtins import str
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.variables.variable import Variable
from fipy.variables.constant import _Constant
from fipy.tools import numerix
from functools import reduce

class _MeshVariable(Variable):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.

    Abstract base class for a `Variable` that is defined on a mesh
    """
    def __init__(self, mesh, name='', value=0., rank=None, elementshape=None,
                 unit=None, cached=1):
        """
        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            the mesh that defines the geometry of this `Variable`
        name : str
            the user-readable name of the `Variable`
        value : float or array_like
            the initial value
        rank : int
            the rank (number of dimensions) of each element of this
            `Variable`. Default: 0
        elementshape : :obj:`tuple` of :obj:`int`
            the shape of each element of this variable
            Default: `rank * (mesh.dim,)`
        unit : str or ~fipy.tools.dimensions.physicalField.PhysicalUnit
            The physical units of the variable
        """
        if isinstance(value, (list, tuple)):
            value = numerix.array(value)

        if isinstance(value, _MeshVariable):
            if mesh is None:
                mesh = value.mesh
            elif mesh != value.mesh:
                raise ValueError("The new 'Variable' must use the same mesh as the supplied value")

        self.mesh = mesh
        value = self._globalToLocalValue(value)

        if value is None:
            array = None
        elif not isinstance(value, _Constant) and isinstance(value, Variable):
            name = name or value.name
            unit = None
            if isinstance(value, _MeshVariable):
                if not isinstance(value, self._variableClass):
                    raise TypeError("A '%s' cannot be cast to a '%s'" % (value._variableClass.__name__,
                                                                          self._variableClass.__name__))
                if elementshape is not None and elementshape != value.shape[:-1]:
                    raise ValueError("'elementshape' != shape of elements of 'value'")

                if rank is not None and rank != value.rank:
                    raise ValueError("'rank' != rank of 'value'")

                elementshape = value.shape[:-1]
                array = None

#             value = value._copyValue()

        if elementshape is None:
            valueShape = numerix.getShape(value)
            if valueShape != () and valueShape[-1] == self._getShapeFromMesh(mesh)[-1]:
                if elementshape is not None and elementshape != valueShape[:-1]:
                    raise ValueError("'elementshape' != shape of elements of 'value'")

                if rank is not None and rank != len(valueShape[:-1]):
                    raise ValueError("'rank' != rank of 'value'")
                elementshape = valueShape[:-1]
            elif rank is None and elementshape is None:
                elementshape = valueShape

        if rank is None:
            if elementshape is None:
                elementshape = ()
        elif elementshape is None:
            elementshape = rank * (mesh.dim,)
        elif len(elementshape) != rank:
            raise ValueError('len(elementshape) != rank')

        self.elementshape = elementshape

        if not "array" in locals():
            if numerix._isPhysical(value):
                dtype = numerix.obj2sctype(value.value)
            else:
                dtype = numerix.obj2sctype(value)
            #print "meshvariable elshape: ",self.elementshape
            #print "meshvariable _getShapeFromMesh: ",self._getShapeFromMesh(mesh)
            array = numerix.zeros(self.elementshape
                                  + self._getShapeFromMesh(mesh),
                                  dtype)
            if numerix._broadcastShape(array.shape, numerix.shape(value)) is None:
                if not isinstance(value, Variable):
                    value = _Constant(value)
                value = value[..., numerix.newaxis]

        Variable.__init__(self, name=name, value=value, unit=unit,
                          array=array, cached=cached)

    def _globalToLocalValue(self, value):
        if value is not None:
            if not isinstance(value, Variable):
                value = _Constant(value)
            valueShape = value.shape
            if valueShape is not () and valueShape[-1] == self._globalNumberOfElements:
                if valueShape[-1] != 0:
                    # workaround for NumPy:ticket:1171
                    value = value[..., self._globalOverlappingIDs]

            value = value.value
        return value

    def _getGlobalValue(self, localIDs, globalIDs):
        localValue = self.value
        if self.mesh.communicator.Nproc > 1:
            if localValue.shape[-1] != 0:
                localValue = localValue[..., localIDs]
            globalIDs = numerix.concatenate(self.mesh.communicator.allgather(globalIDs))

            globalValue = numerix.empty(localValue.shape[:-1] + (max(globalIDs) + 1,),
                                        dtype=numerix.obj2sctype(localValue))
            globalValue[..., globalIDs] = numerix.concatenate(self.mesh.communicator.allgather(localValue), axis=-1)

            return globalValue
        else:
            return localValue

    def __str__(self):
        return str(self.globalValue)

    def __repr__(self):
        if hasattr(self, 'name') and len(self.name) > 0:
            return self.name
        else:
            s = self.__class__.__name__ + '('
            s += 'value=' + repr(self.globalValue)
            s += ')'
            if len(self.name) == 0:
                s = s[:-1] + ', mesh=' + repr(self.mesh) + s[-1]
            return s

    @property
    def constraintMask(self):
        r"""
        Test that `constraintMask` returns a Variable that updates itself whenever the constraints change.

        >>> from fipy import *

        >>> m = Grid2D(nx=2, ny=2)
        >>> x, y = m.cellCenters
        >>> v0 = CellVariable(mesh=m)
        >>> v0.constrain(1., where=m.facesLeft)
        >>> print(v0.faceValue.constraintMask)
        [False False False False False False  True False False  True False False]
        >>> print(v0.faceValue)
        [ 0.  0.  0.  0.  0.  0.  1.  0.  0.  1.  0.  0.]
        >>> v0.constrain(3., where=m.facesRight)
        >>> print(v0.faceValue.constraintMask)
        [False False False False False False  True False  True  True False  True]
        >>> print(v0.faceValue)
        [ 0.  0.  0.  0.  0.  0.  1.  0.  3.  1.  0.  3.]
        >>> v1 = CellVariable(mesh=m)
        >>> v1.constrain(1., where=(x < 1) & (y < 1))
        >>> print(v1.constraintMask)
        [ True False False False]
        >>> print(v1)
        [ 1.  0.  0.  0.]
        >>> v1.constrain(3., where=(x > 1) & (y > 1))
        >>> print(v1.constraintMask)
        [ True False False  True]
        >>> print(v1)
        [ 1.  0.  0.  3.]

        """
        if not hasattr(self, '_constraintMask'):
            from fipy.variables.constraintMask import _ConstraintMask
            self._constraintMask = _ConstraintMask(self)
        return self._constraintMask

    def constrain(self, value, where=None):
        super(_MeshVariable, self).constrain(value, where=where)
        if hasattr(self, '_constraintMask'):
            self._constraintMask._requires(self._constraints[-1].where)

    def _getShapeFromMesh(mesh):
        """
        Return the shape of this `MeshVariable` type, given a particular mesh.
        Return `None` if unknown or independent of the mesh.
        """
        return None

    _getShapeFromMesh = staticmethod(_getShapeFromMesh)

    @property
    def shape(self):
        """
        >>> from fipy.meshes import Grid2D
        >>> from fipy.variables.cellVariable import CellVariable
        >>> mesh = Grid2D(nx=2, ny=3)
        >>> var = CellVariable(mesh=mesh)
        >>> print(numerix.allequal(var.shape, (6,))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allequal(var.arithmeticFaceValue.shape, (17,))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allequal(var.grad.shape, (2, 6))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allequal(var.faceGrad.shape, (2, 17))) # doctest: +PROCESSOR_0
        True

        Have to account for zero length arrays

        >>> from fipy import Grid1D
        >>> m = Grid1D(nx=0)
        >>> v = CellVariable(mesh=m, elementshape=(2,))
        >>> numerix.allequal((v * 1).shape, (2, 0))
        True

        """
        return (Variable._getShape(self)
                or (self.elementshape + self._getShapeFromMesh(self.mesh))
                or ())

    def _dot(a, b, index):
        """
        Workhorse method to calculate the scalar product

        .. math::

           \mathsf{a} \cdot \mathsf{b}

        for all but the last index of `a` and `b`. Both `a` and `b` can be of
        arbitrary rank, but at this point, both must be appropriately broadcast
        `array` objects.
        """
        rankA = len(a.shape) - 1
        rankB = len(b.shape) - 1
        if rankA <= 0 or rankB <= 0:
            # if either a or b is scalar, then just do multiplication
            return a[index] * b
        else:
            return numerix.sum(a[index] * b, axis=rankA - 1)
    _dot = staticmethod(_dot)

    def __dot(A, B, operatorClass):
        """
        Workhorse method to return a `_BinaryOperatorVariable` that will
        dynamically perform the mesh-element--by--mesh-element (cell-by-cell,
        face-by-face, etc.) scalar product

        .. math::

           \mathsf{A} \cdot \mathsf{B}

        Both `A` and `B` can be of arbitrary rank, but at this point, both must
        be appropriately broadcast `_MeshVariable` objects.

        Test for inline bug

        >>> from fipy import *
        >>> m = Grid2D(nx=2, ny=2)
        >>> v = FaceVariable(mesh=m, rank=1, value=m._orientedFaceNormals)

        >>> print(len(v.dot(1.).shape))
        2
        >>> print(numerix.allequal(v.dot(1.).globalValue.shape, (2, 12)))
        True
        >>> tmp = m._cellDistances * v.dot(1.)
        >>> print(numerix.allequal(tmp.globalValue.shape, (2, 12)))
        True

        The value shouldn't change shape the second time it's
        evaluated. The second time is inline and the inline code does
        not have the correct shape.

        >>> print(numerix.allequal(tmp.globalValue.shape, (2, 12)))
        True

        More inconsistent shape problems.

        >>> m = Grid2D(nx=3, ny=3)
        >>> v0 = FaceVariable(mesh=m, rank=1, value=m._orientedFaceNormals)
        >>> print(len(v0.dot(m.faceCenters[0]).shape))
        2
        >>> print(numerix.allequal(v0.dot(m.faceCenters[0]).globalValue.shape, (2, 24)))
        True

        """
        rankA = len(A.shape) - 1
        rankB = len(B.shape) - 1

        index = (numerix.index_exp[...] + (numerix.newaxis,) * (rankB - 1)
                 + numerix.index_exp[:])
        opShape = numerix._broadcastShape(A[index].shape, B.shape)

        if rankA > 0 and rankB > (rankA - 1):
            opShape = opShape[:rankA-1] + opShape[rankA:]

        return A._BinaryOperatorVariable(lambda a, b: _MeshVariable._dot(a, b, index),
                                         B,
                                         opShape=opShape,
                                         operatorClass=operatorClass,
                                         canInline=False)
    __dot = staticmethod(__dot)

    def dot(self, other, opShape=None, operatorClass=None):
        """
        Return the mesh-element--by--mesh-element (cell-by-cell, face-by-face,
        etc.) scalar product

        .. math::

           \text{self} \cdot \text{other}

        Both `self` and `other` can be of arbitrary rank, and `other` does not
        need to be a `_MeshVariable`.
        """
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)
        opShape, baseClass, other = self._shapeClassAndOther(opShape=None, operatorClass=None, other=other)

        return _MeshVariable.__dot(self, other, self._OperatorVariableClass(baseClass))

    def rdot(self, other, opShape=None, operatorClass=None):
        """
        Return the mesh-element--by--mesh-element (cell-by-cell, face-by-face,
        etc.) scalar product

        .. math::

           \text{other} \cdot \text{self}

        Both `self` and `other` can be of arbitrary rank, and `other` does not
        need to be a `_MeshVariable`.
        """
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)
        opShape, baseClass, other = self._shapeClassAndOther(opShape=None, operatorClass=None, other=other)

        return self.__dot(other, self, self._OperatorVariableClass(baseClass))

    def _maxminparallel_(self, a, axis, default, fn, fnParallel):
        a = a[..., self._localNonOverlappingIDs]

        if numerix.multiply.reduce(a.shape) == 0:
            if axis is None:
                opShape = ()
            else:
                opShape=self.shape[:axis] + self.shape[axis+1:]

            if len(opShape) == 0:
                nodeVal = default
            else:
                nodeVal = numerix.empty(opShape)
                nodeVal[:] = default
        else:
            nodeVal = fn(axis=axis)

        return fnParallel(nodeVal)

    def max(self, axis=None):
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def maxParallel(a):
                return self._maxminparallel_(a=a, axis=axis, default=-numerix.inf,
                                             fn=a.max, fnParallel=self.mesh.communicator.MaxAll)

            return self._axisOperator(opname="maxVar",
                                      op=maxParallel,
                                      axis=axis)
        else:
            return Variable.max(self, axis=axis)

    def min(self, axis=None):
        """
        >>> from fipy import Grid2D, CellVariable
        >>> mesh = Grid2D(nx=5, ny=5)
        >>> x, y = mesh.cellCenters
        >>> v = CellVariable(mesh=mesh, value=x*y)
        >>> print(v.min())
        0.25
        """
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def minParallel(a):
                return self._maxminparallel_(a=a, axis=axis, default=numerix.inf,
                                             fn=a.min, fnParallel=self.mesh.communicator.MinAll)

            return self._axisOperator(opname="minVar",
                                      op=minParallel,
                                      axis=axis)
        else:
            return Variable.min(self, axis=axis)

    def all(self, axis=None):
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def allParallel(a):
                a = a[..., self._localNonOverlappingIDs]
                return self.mesh.communicator.all(a, axis=axis)

            return self._axisOperator(opname="allVar",
                                      op=allParallel,
                                      axis=axis)
        else:
            return Variable.all(self, axis=axis)

    def any(self, axis=None):
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def anyParallel(a):
                a = a[..., self._localNonOverlappingIDs]
                return self.mesh.communicator.any(a, axis=axis)

            return self._axisOperator(opname="anyVar",
                                      op=anyParallel,
                                      axis=axis)
        else:
            return Variable.any(self, axis=axis)

    def sum(self, axis=None):
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def sumParallel(a):
                a = a[..., self._localNonOverlappingIDs]
                return self.mesh.communicator.sum(a, axis=axis)

            return self._axisOperator(opname="sumVar",
                                      op=sumParallel,
                                      axis=axis)
        else:
            return Variable.sum(self, axis=axis)

    def allclose(self, other, rtol=1.e-5, atol=1.e-8):
         if self.mesh.communicator.Nproc > 1:
             def allcloseParallel(a, b):
                 return self.mesh.communicator.allclose(a, b, rtol=rtol, atol=atol)

             operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
             return self._BinaryOperatorVariable(allcloseParallel,
                                                 other,
                                                 operatorClass=operatorClass,
                                                 opShape=(),
                                                 canInline=False)
         else:
             return Variable.allclose(self, other, rtol=rtol, atol=atol)

    def allequal(self, other):
         if self.mesh.communicator.Nproc > 1:
             def allequalParallel(a, b):
                 return self.mesh.communicator.allequal(a, b)

             operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
             return self._BinaryOperatorVariable(allequalParallel,
                                                 other,
                                                 operatorClass=operatorClass,
                                                 opShape=(),
                                                 canInline=False)
         else:
             return Variable.allequal(self, other)

    def std(self, axis=None, **kwargs):
        """Evaluate standard deviation of all the elements of a `MeshVariable`.

        Adapted from http://mpitutorial.com/tutorials/mpi-reduce-and-allreduce/

        >>> import fipy as fp
        >>> mesh = fp.Grid2D(nx=2, ny=2, dx=2., dy=5.)
        >>> var = fp.CellVariable(value=(1., 2., 3., 4.), mesh=mesh)
        >>> print((var.std()**2).allclose(1.25))
        True
        """
        if self.mesh.communicator.Nproc > 1 and (axis is None or axis == len(self.shape) - 1):
            def stdParallel(a):
                N = self.mesh.globalNumberOfCells
                mean = self.sum(axis=axis).value / N
                sq_diff = (self - mean)**2

                return numerix.sqrt(sq_diff.sum(axis=axis).value / N)

            return self._axisOperator(opname="stdVar",
                                      op=stdParallel,
                                      axis=axis)
        else:
            return Variable.std(self, axis=axis)

    def _shapeClassAndOther(self, opShape, operatorClass, other):
        """
        Determine the shape of the result, the base class of the result, and (if
        necessary) a modified form of `other` that is suitable for the
        operation.

        By default, returns the result of the generic
        `Variable._shapeClassAndOther()`, but if that fails, and if each
        dimension of `other` is exactly the `Mesh` dimension, do what the user
        probably "meant" and project `other` onto the `Mesh`.

        >>> from fipy import *
        >>> mesh = Grid1D(nx=5)
        >>> A = numerix.arange(5)
        >>> B = Variable(1.)
        >>> import warnings
        >>> savedFilters = list(warnings.filters)
        >>> warnings.resetwarnings()
        >>> warnings.simplefilter("error", UserWarning, append=True)
        >>> C = (A * B) * CellVariable(mesh=mesh)
        Traceback (most recent call last):
          ...
        UserWarning: The expression `(Variable(value=array(1.0)) * [0 1 2 3 4])` has been cast to a constant `CellVariable`
        >>> warnings.filters = savedFilters
        """
        otherShape = numerix.getShape(other)
        if (not isinstance(other, _MeshVariable)
            and otherShape is not ()
            and otherShape[-1] == self._globalNumberOfElements):
            if (isinstance(other, Variable) and len(other.requiredVariables) > 0):
                import warnings
                warnings.warn("The expression `%s` has been cast to a constant `%s`"
                              % (repr(other), self._variableClass.__name__),
                              UserWarning, stacklevel=4)
            other = self._variableClass(value=other, mesh=self.mesh)

        newOpShape, baseClass, newOther = Variable._shapeClassAndOther(self, opShape, operatorClass, other)

        if ((newOpShape is None or baseClass is None)
            and numerix.alltrue(numerix.array(numerix.getShape(other)) == self.mesh.dim)):
                newOpShape, baseClass, newOther = Variable._shapeClassAndOther(self, opShape, operatorClass, other[..., numerix.newaxis])

        return (newOpShape, baseClass, newOther)

    def _OperatorVariableClass(self, baseClass=None):
        baseClass = Variable._OperatorVariableClass(self, baseClass=baseClass)

        class _MeshOperatorVariable(baseClass):
            def __init__(self, op, var, opShape=None, canInline=True,
                         *args, **kwargs):
                mesh = reduce(lambda a, b: a or b,
                              [getattr(v, "mesh", None) for v in var])
                for shape in [opShape] + [getattr(v, "opShape", None) for v in var]:
                    if shape is not None:
                        opShape = shape
                        break
##                 opShape = reduce(lambda a, b: a or b,
##                                  [opShape] + [getattr(v, "opShape", None) for v in var])
                if opShape is not None:
                    elementshape = opShape[:-1]
                else:
                    elementshape = reduce(lambda a, b: a or b,
                                          [getattr(v, "elementshape", None) for v in var])

                baseClass.__init__(self, mesh=mesh, op=op, var=var,
                                   opShape=opShape, canInline=canInline,
                                   elementshape=elementshape,
                                   *args, **kwargs)

            @property
            def rank(self):
                return len(self.opShape) - 1

        return _MeshOperatorVariable

    @property
    def rank(self):
        return len(self.shape) - 1

    def setValue(self, value, unit = None, where = None):
        if where is not None:
            shape = numerix.getShape(where)
            if shape != self.shape \
              and shape == self._getShapeFromMesh(mesh=self.mesh):
                for dim in self.elementshape:
                    where = numerix.repeat(where[numerix.newaxis, ...], repeats=dim, axis=0)

        return Variable.setValue(self, value=value, unit=unit, where=where)

    def _axisClass(self, axis):
        """
        if we operate along the mesh elements, then this is no longer a
        `_MeshVariable`, otherwise we get back a `_MeshVariable` of the same
        class, but lower rank.
        """
        if axis is None or axis == len(self.shape) - 1 or axis == -1:
            return Variable._OperatorVariableClass(self, baseClass=Variable)
        else:
            return self._OperatorVariableClass()

    def _getitemClass(self, index):
        if not isinstance(index, tuple):
            if isinstance(index, list):
                index = tuple(index)
            else:
                index = (index,)
        indexshape = numerix._indexShape(index=index, arrayShape=self.shape)
        if (len(indexshape) > 0
            and indexshape[-1] == self.shape[-1]
            and numerix.obj2sctype(index[-1]) != numerix.obj2sctype(bool)):
            return self._OperatorVariableClass()
        else:
            return Variable._OperatorVariableClass(self, baseClass=Variable)

    def __getstate__(self):
        """
        Used internally to collect the necessary information to ``pickle`` the
        `_MeshVariable` to persistent storage.
        """
        return {
            'mesh': self.mesh,
            'name': self.name,
            'value': self.value,
            'unit': self.unit,
        }


def _testDot(self):
    """
    >>> from fipy import *
    >>> mesh = Grid2D(nx=2, ny=3)

    >>> s1 = CellVariable(mesh=mesh, value=2)
    >>> s2 = CellVariable(mesh=mesh, value=3)

    >>> v1 = CellVariable(mesh=mesh, rank=1,
    ...                   value=numerix.array([2, 3])[..., numerix.newaxis])
    >>> v2 = CellVariable(mesh=mesh, rank=1,
    ...                   value=numerix.array([3, 4])[..., numerix.newaxis])

    >>> t21 = CellVariable(mesh=mesh, rank=2,
    ...                    value=numerix.array([[2, 3],
    ...                                         [4, 5]])[..., numerix.newaxis])
    >>> t22 = CellVariable(mesh=mesh, rank=2,
    ...                    value=numerix.array([[3, 4],
    ...                                         [5, 6]])[..., numerix.newaxis])

    >>> t31 = CellVariable(mesh=mesh, rank=3,
    ...                    value=numerix.array([[[3, 4],
    ...                                         [5, 6]],
    ...                                        [[5, 6],
    ...                                         [7, 8]]])[..., numerix.newaxis])
    >>> t32 = CellVariable(mesh=mesh, rank=3,
    ...                    value=numerix.array([[[2, 3],
    ...                                         [4, 5]],
    ...                                        [[4, 5],
    ...                                         [6, 7]]])[..., numerix.newaxis])

    >>> def P(a):
    ...     a = a.globalValue
    ...     print(a[..., 0], tuple(numerix.asarray(a.shape, dtype='int32')))

    >>> P(v1.dot(v2))
    18 (6,)
    >>> P(v1.dot(t22))
    [21 26] (2, 6)
    >>> P(v1.dot(t31))
    [[21 26]
     [31 36]] (2, 2, 6)

    >>> P(t21.dot(v1))
    [13 23] (2, 6)
    >>> P(t21.dot(t22))
    [[21 26]
     [37 46]] (2, 2, 6)
    >>> P(t21.dot(t31))
    [[[21 26]
      [31 36]]
    <BLANKLINE>
     [[37 46]
      [55 64]]] (2, 2, 2, 6)

    >>> P(t31.dot(v1))
    [[18 28]
     [28 38]] (2, 2, 6)
    >>> P(t31.dot(t21))
    [[[22 29]
      [34 45]]
    <BLANKLINE>
     [[34 45]
      [46 61]]] (2, 2, 2, 6)
    >>> P(t31.dot(t32))
    [[[[22 29]
       [36 43]]
    <BLANKLINE>
      [[34 45]
       [56 67]]]
    <BLANKLINE>
    <BLANKLINE>
     [[[34 45]
       [56 67]]
    <BLANKLINE>
      [[46 61]
       [76 91]]]] (2, 2, 2, 2, 6)

    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.diffusionTermNoCorrection import DiffusionTermNoCorrection
from fipy.terms.diffusionTermCorrection import DiffusionTermCorrection
from fipy.tools import numerix

__all__ = ["DiffusionTerm", "DiffusionTermCorrection", "DiffusionTermNoCorrection"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class DiffusionTerm(DiffusionTermNoCorrection):
    r"""

    This term represents a higher order diffusion term. The order of the term is determined
    by the number of `coeffs`, such that::

        DiffusionTerm(D1)

    represents a typical 2nd-order diffusion term of the form

    .. math::

       \nabla\cdot\left(D_1 \nabla \phi\right)

    and::

        DiffusionTerm((D1,D2))

    represents a 4th-order Cahn-Hilliard term of the form

    .. math::

       \nabla \cdot \left\{ D_1 \nabla \left[ \nabla\cdot\left( D_2 \nabla \phi\right) \right] \right\}

    and so on.

    """

    def _test(self):
        r"""
        Test, 2nd order, 1 dimension, fixed flux of zero both ends.

        >>> from fipy.meshes import Grid1D
        >>> from fipy.solvers import _MeshMatrix
        >>> from fipy.variables.cellVariable import CellVariable
        >>> mesh = Grid1D(dx = 1., nx = 2)
        >>> term = DiffusionTerm(coeff = (1,))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 1., -1.),
        ...                         (-1.,  1.)))) # doctest: +PROCESSOR_0
        True
        >>> from fipy.variables.cellVariable import CellVariable
        >>> v, L, b = term._buildMatrix(var=CellVariable(mesh=mesh), SparseMatrix=_MeshMatrix)
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -1.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (0., 0.))) # doctest: +PROCESSOR_0
        True

        The coefficient must be a `FaceVariable`, a `CellVariable` (which will
        be interpolated to a `FaceVariable`), or a scalar value

        >>> from fipy.variables.faceVariable import FaceVariable
        >>> term = DiffusionTerm(coeff=FaceVariable(mesh=mesh, value=1))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A, (( 1., -1.),
        ...                            (-1.,  1.)))) # doctest: +PROCESSOR_0
        True
        >>> v, L, b = term._buildMatrix(var=CellVariable(mesh=mesh), SparseMatrix=_MeshMatrix)
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -1.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (0., 0.))) # doctest: +PROCESSOR_0
        True

        >>> term = DiffusionTerm(coeff=CellVariable(mesh=mesh, value=1))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 1., -1.),
        ...                         (-1.,  1.)))) # doctest: +PROCESSOR_0
        True
        >>> v, L, b = term._buildMatrix(var=CellVariable(mesh=mesh), SparseMatrix=_MeshMatrix)
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -1.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (0., 0.))) # doctest: +PROCESSOR_0
        True

        >>> from fipy.variables.variable import Variable
        >>> term = DiffusionTerm(coeff = Variable(value = 1))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 1., -1.),
        ...                         (-1.,  1.)))) # doctest: +PROCESSOR_0
        True
        >>> v, L, b = term._buildMatrix(var=CellVariable(mesh=mesh), SparseMatrix=_MeshMatrix)
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -1.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (0., 0.))) # doctest: +PROCESSOR_0
        True

        >>> term = DiffusionTerm(coeff = ((1, 2),))

        >>> term = DiffusionTerm(coeff = FaceVariable(mesh = mesh, value = (1,), rank=1))
        >>> term = DiffusionTerm(coeff = CellVariable(mesh=mesh, value=(1,), rank=1))

        Test, 2nd order, 1 dimension, fixed flux 3, fixed value of 4

        >>> var=CellVariable(mesh=mesh)
        >>> var.faceGrad.constrain([-3.], mesh.facesLeft)
        >>> var.constrain(4., mesh.facesRight)
        >>> term = DiffusionTerm(coeff = (1.,))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 1., -1.),
        ...                         (-1.,  1.)))) # doctest: +PROCESSOR_0
        True
        >>> v, L, b = term._buildMatrix(var=var,
        ...                         SparseMatrix=_MeshMatrix)
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -3.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (-3., -8.))) # doctest: +PROCESSOR_0
        True

        Test, 4th order, 1 dimension, x = 0; fixed flux 3, fixed curvatures 0,
        x = 2, fixed value 1, fixed curvature 0

        >>> from fipy.boundaryConditions.nthOrderBoundaryCondition \
        ...     import NthOrderBoundaryCondition
        >>> bcLeft2 =  NthOrderBoundaryCondition(mesh.facesLeft, 0., 2)
        >>> var = CellVariable(mesh=mesh)
        >>> var.faceGrad.constrain([-3.], mesh.facesLeft)
        >>> var.constrain(4., mesh.facesRight)
        >>> bcRight2 =  NthOrderBoundaryCondition(mesh.facesRight, 0., 2)
        >>> term = DiffusionTerm(coeff = (1., 1.))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 1., -1.),
        ...                         (-1.,  1.)))) # doctest: +PROCESSOR_0
        True

        >>> v, L, b = term._buildMatrix(var=var, SparseMatrix=_MeshMatrix,
        ...                         boundaryConditions=(bcLeft2, bcRight2))
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 4., -6.),
        ...                         (-4., 10.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (1., 21.))) # doctest: +PROCESSOR_0
        True

        Test, 4th order, 1 dimension, x = 0; fixed flux 3, fixed curvature 2,
        x = 2, fixed value 4, fixed 3rd order `-1`

        >>> bcLeft2 =  NthOrderBoundaryCondition(mesh.facesLeft, 2., 2)
        >>> bcRight2 =  NthOrderBoundaryCondition(mesh.facesRight, -1., 3)

        >>> var = CellVariable(mesh=mesh)
        >>> var.faceGrad.constrain([-3.], mesh.facesLeft)
        >>> var.constrain(4., mesh.facesRight)

        >>> term = DiffusionTerm(coeff = (-1., 1.))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-1.,  1.),
        ...                         ( 1., -1.)))) # doctest: +PROCESSOR_0
        True

        >>> v, L, b = term._buildMatrix(var=var,
        ...                         SparseMatrix=_MeshMatrix,
        ...                         boundaryConditions = (bcLeft2, bcRight2))
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        ((-4.,  6.),
        ...                         ( 2., -4.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (3., -4.))) # doctest: +PROCESSOR_0
        True


        Test when `dx = 0.5`.

        >>> mesh = Grid1D(dx = .5, nx = 2)

        >>> bcLeft2 =  NthOrderBoundaryCondition(mesh.facesLeft, 1., 2)
        >>> bcRight2 =  NthOrderBoundaryCondition(mesh.facesRight, 0., 3)

        >>> var = CellVariable(mesh=mesh)
        >>> var.faceGrad.constrain([1.], mesh.facesRight)
        >>> var.constrain(0., mesh.facesLeft)

        >>> term = DiffusionTerm(coeff = (1., 1.))
        >>> coeff = term._getGeomCoeff(CellVariable(mesh=mesh))
        >>> M = term._getCoefficientMatrixForTests(_MeshMatrix, CellVariable(mesh=mesh), coeff[0])
        >>> A = M.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 2., -2.),
        ...                         (-2.,  2.)))) # doctest: +PROCESSOR_0
        True

        >>> v, L, b = term._buildMatrix(var=var,
        ...                         SparseMatrix=_MeshMatrix,
        ...                         boundaryConditions = (bcLeft2, bcRight2))
        >>> A = L.numpyArray
        >>> print(numerix.allclose(A,
        ...                        (( 80., -32.),
        ...                         (-32.,  16.)))) # doctest: +PROCESSOR_0
        True
        >>> print(numerix.allclose(b, (-8., 4.))) # doctest: +PROCESSOR_0
        True

        The following tests are to check that `DiffusionTerm` can take any of the four
        main Variable types.

        >>> from fipy.meshes.tri2D import Tri2D
        >>> mesh = Tri2D(nx = 1, ny = 1)
        >>> term = DiffusionTermCorrection(CellVariable(value = 1, mesh = mesh))
        >>> print(term._getGeomCoeff(CellVariable(mesh=mesh))[0])
        [ 6.   6.   6.   6.   1.5  1.5  1.5  1.5]
        >>> term = DiffusionTerm(FaceVariable(value = 1, mesh = mesh))
        >>> print(term._getGeomCoeff(CellVariable(mesh=mesh))[0])
        [ 6.   6.   6.   6.   1.5  1.5  1.5  1.5]
        >>> term = DiffusionTerm(CellVariable(value=((0.5,), (1,)), mesh=mesh, rank=1))
        >>> print(term._getGeomCoeff(CellVariable(mesh=mesh))[0])
        [ 6.     6.     3.     3.     1.125  1.125  1.125  1.125]
        >>> term = DiffusionTerm(FaceVariable(value=((0.5,), (1,)), mesh=mesh, rank=1))
        >>> print(term._getGeomCoeff(CellVariable(mesh=mesh))[0])
        [ 6.     6.     3.     3.     1.125  1.125  1.125  1.125]
        >>> mesh = Tri2D(nx = 1, ny = 1, dy = 0.1)
        >>> term = DiffusionTermCorrection(FaceVariable(value=((0.5,), (1,)), mesh=mesh, rank=1))
        >>> val = (60., 60., 0.3, 0.3, 0.22277228, 0.22277228, 0.22277228, 0.22277228)
        >>> print(numerix.allclose(term._getGeomCoeff(CellVariable(mesh=mesh))[0], val))
        1
        >>> term = DiffusionTerm((((0.5,), (1,)),))
        >>> print(numerix.allclose(term._getGeomCoeff(CellVariable(mesh=mesh))[0], val))
        Traceback (most recent call last):
            ...
        IndexError: diffusion coefficient tensor is not an appropriate shape for this mesh

        Anisotropy test

        >>> from fipy.meshes.tri2D import Tri2D
        >>> mesh = Tri2D(nx = 1, ny = 1)
        >>> term = DiffusionTerm((((1, 2), (3, 4)),))
        >>> print(numerix.allclose(term._getGeomCoeff(CellVariable(mesh=mesh)),
        ...                        [[ 24.,         24.,          6.,          6.,          0.,          7.5,
        ...                        7.5,         0.        ],
        ...                         [ -3.,         -3.,          2.,          2.,         -1.41421356,
        ...                        0.70710678,   0.70710678,  -1.41421356]]))
        True

        Negate the term.

        >>> -DiffusionTerm(coeff=[1.])
        DiffusionTerm(coeff=[-1.0])

        >>> -DiffusionTerm()
        DiffusionTerm(coeff=[-1.0])

        Testing vector diffusion terms by comparing with coupled solutions.

        >>> from fipy import *
        >>> m = Grid1D(nx=6)
        >>> q = CellVariable(mesh=m, elementshape=(2,))
        >>> coeff = FaceVariable(mesh=m, elementshape=(2, 2), value=[[1, 2], [3, 4]])
        >>> vectorEq = DiffusionTerm(coeff) == 0
        >>> vectorEq.cacheMatrix()
        >>> vectorEq.solve(q, solver=DummySolver())

        >>> v0 = CellVariable(mesh=m)
        >>> v1 = CellVariable(mesh=m)
        >>> coupledEq = ((DiffusionTerm(coeff=1., var=v0) + DiffusionTerm(coeff=2., var=v1)) & (DiffusionTerm(coeff=3., var=v0) + DiffusionTerm(coeff=4., var=v1)))((v0, v1))
        >>> coupledEq.cacheMatrix()
        >>> coupledEq.solve(solver=DummySolver())
        >>> print((coupledEq.matrix.numpyArray == vectorEq.matrix.numpyArray).all())
        True

        Test vector diffusion terms.

        >>> m = Grid2D(nx=2, ny=2)
        >>> q = CellVariable(mesh=m, elementshape=(3,))
        >>> coeff = FaceVariable(mesh=m, elementshape=(3, 3), value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        >>> vectorEq0 = DiffusionTerm(coeff) == 0
        >>> vectorEq0.cacheMatrix()
        >>> vectorEq0.solve(q, solver=DummySolver())

        Test vector diffusion terms when the coefficient is a tuple or list.

        >>> m = Grid2D(nx=2, ny=2)
        >>> q = CellVariable(mesh=m, elementshape=(3,))
        >>> vectorEq1 = DiffusionTerm(([[1, 2, 3], [4, 5, 6], [7, 8, 9]],)) == 0
        >>> vectorEq1.cacheMatrix()
        >>> vectorEq1.solve(q, solver=DummySolver())

        The following test tests for negating a tuple/list coefficient for the diffusion term.

        >>> m = Grid2D(nx=2, ny=2)
        >>> q = CellVariable(mesh=m, elementshape=(3,))
        >>> vectorEq2 = TransientTerm(1e-20) == DiffusionTerm(([[1, 2, 3], [4, 5, 6], [7, 8, 9]],))
        >>> vectorEq2.cacheMatrix()
        >>> vectorEq2.solve(q, solver=DummySolver(), dt=1.)

        Test the previous three vector examples against coupled.

        >>> v0 = CellVariable(mesh=m)
        >>> v1 = CellVariable(mesh=m)
        >>> v2 = CellVariable(mesh=m)
        >>> coupledEq = ((DiffusionTerm(coeff=1., var=v0) + DiffusionTerm(coeff=2., var=v1) + DiffusionTerm(coeff=3., var=v2)) & \
        ...              (DiffusionTerm(coeff=4., var=v0) + DiffusionTerm(coeff=5., var=v1) + DiffusionTerm(coeff=6., var=v2)) & \
        ...              (DiffusionTerm(coeff=7., var=v0) + DiffusionTerm(coeff=8., var=v1) + DiffusionTerm(coeff=9, var=v2)))((v0, v1, v2))
        >>> coupledEq.cacheMatrix()
        >>> coupledEq.solve(solver=DummySolver())
        >>> coupledMatrix = coupledEq.matrix.numpyArray
        >>> print((coupledMatrix == vectorEq0.matrix.numpyArray).all())
        True
        >>> print((coupledMatrix == vectorEq1.matrix.numpyArray).all())
        True
        >>> print((coupledMatrix == -vectorEq2.matrix.numpyArray).all())
        True

        """
        pass



def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import division
from __future__ import unicode_literals
from builtins import range
__docformat__ = 'restructuredtext'

from fipy.meshes.abstractMesh import AbstractMesh
from fipy.meshes.representations.meshRepresentation import _MeshRepresentation
from fipy.meshes.topologies.meshTopology import _MeshTopology

from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import serialComm

__all__ = ["MeshAdditionError", "Mesh"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MeshAdditionError(Exception):
    pass

class Mesh(AbstractMesh):
    """Generic mesh class using numerix to do the calculations

        Meshes contain cells, faces, and vertices.

        This is built for a non-mixed element mesh.
    """

    def __init__(self, vertexCoords, faceVertexIDs, cellFaceIDs, communicator=serialComm, _RepresentationClass=_MeshRepresentation, _TopologyClass=_MeshTopology):
        super(Mesh, self).__init__(communicator=communicator,
                                   _RepresentationClass=_RepresentationClass,
                                   _TopologyClass=_TopologyClass)

        """faceVertexIds and cellFacesIds must be padded with minus ones."""

        self.vertexCoords = vertexCoords
        self.faceVertexIDs = MA.masked_values(faceVertexIDs, -1)
        self.cellFaceIDs = MA.masked_values(cellFaceIDs, -1)

        self.dim = self.vertexCoords.shape[0]

        if not hasattr(self, "numberOfFaces"):
            self.numberOfFaces = self.faceVertexIDs.shape[-1]
        if not hasattr(self, "numberOfCells"):
            self.numberOfCells = self.cellFaceIDs.shape[-1]
        if not hasattr(self, "globalNumberOfCells"):
            self.globalNumberOfCells = self.numberOfCells
        if not hasattr(self, "globalNumberOfFaces"):
            self.globalNumberOfFaces = self.numberOfFaces

        self.faceCellIDs = self._calcFaceCellIDs()

        self._setTopology()
        self._setGeometry(scaleLength = 1.)

    """
    Topology set and calculate
    """

    def _setTopology(self):
        (self._interiorFaces,
         self._exteriorFaces) = self._calcInteriorAndExteriorFaceIDs()
        (self._interiorCellIDs,
         self._exteriorCellIDs) = self._calcInteriorAndExteriorCellIDs()
        self._cellToFaceOrientations = self._calcCellToFaceOrientations()
        self._adjacentCellIDs = self._calcAdjacentCellIDs()
        self._cellToCellIDs = self._calcCellToCellIDs()
        self._cellToCellIDsFilled = self._calcCellToCellIDsFilled()

    def _calcInteriorAndExteriorFaceIDs(self):
        from fipy.variables.faceVariable import FaceVariable
        mask = MA.getmask(self.faceCellIDs[1])
        exteriorFaces = FaceVariable(mesh=self,
                                     value=mask)
        interiorFaces = FaceVariable(mesh=self,
                                     value=numerix.logical_not(mask))
        return interiorFaces, exteriorFaces

    def _calcInteriorAndExteriorCellIDs(self):
        try:
            import sys
            if sys.version_info < (2, 6):
                from sets import Set as set
            exteriorCellIDs = set(self.faceCellIDs[0, self._exteriorFaces.value])
            interiorCellIDs = list(set(range(self.numberOfCells)) - self._exteriorCellIDs)
            exteriorCellIDs = list(self._exteriorCellIDs)
        except:
            exteriorCellIDs = self.faceCellIDs[0, self._exteriorFaces.value]
            tmp = numerix.zeros(self.numberOfCells, 'l')
            numerix.put(tmp, exteriorCellIDs, numerix.ones(len(exteriorCellIDs), 'l'))
            exteriorCellIDs = numerix.nonzero(tmp)
            interiorCellIDs = numerix.nonzero(numerix.logical_not(tmp))
        return interiorCellIDs, exteriorCellIDs

    def _calcCellToFaceOrientations(self):
        tmp = numerix.take(self.faceCellIDs[0], self.cellFaceIDs)
        return (tmp == MA.indices(tmp.shape)[-1]) * 2 - 1

    def _calcAdjacentCellIDs(self):
        return (MA.filled(self.faceCellIDs[0]),
                          MA.filled(MA.where(MA.getmaskarray(self.faceCellIDs[1]),
                              self.faceCellIDs[0],
                                             self.faceCellIDs[1])))

    def _calcCellToCellIDs(self):
        cellToCellIDs = numerix.take(self.faceCellIDs, self.cellFaceIDs, axis=1)
        cellToCellIDs = MA.where(self._cellToFaceOrientations == 1,
                                 cellToCellIDs[1], cellToCellIDs[0])
        return cellToCellIDs

    def _calcCellToCellIDsFilled(self):
        N = self.numberOfCells
        M = self._maxFacesPerCell
        cellIDs = numerix.repeat(numerix.arange(N)[numerix.newaxis, ...], M, axis=0)
        return MA.where(MA.getmaskarray(self._cellToCellIDs), cellIDs,
                        self._cellToCellIDs)

    """
    Geometry set and calculate
    """

    def _setGeometry(self, scaleLength = 1.):
        self._faceCenters = self._calcFaceCenters()
        self._faceAreas = self._calcFaceAreas()
        self._cellCenters = self._calcCellCenters()
        (self._internalFaceToCellDistances,
         self._cellToFaceDistanceVectors) = self._calcFaceToCellDistAndVec()
        (self._internalCellDistances,
         self._cellDistanceVectors) = self._calcCellDistAndVec()
        self.faceNormals = self._calcFaceNormals()
        self._orientedFaceNormals = self._calcOrientedFaceNormals()
        self._cellVolumes = self._calcCellVolumes()
        self._faceCellToCellNormals = self._calcFaceCellToCellNormals()
        (self._faceTangents1,
         self._faceTangents2) = self._calcFaceTangents()
        self._cellToCellDistances = self._calcCellToCellDist()

        self._setScaledGeometry(self.scale['length'])

        self._cellAreas = self._calcCellAreas()
        self._cellNormals = self._calcCellNormals()

    def _calcFaceAreas(self):
        faceVertexIDs = MA.filled(self.faceVertexIDs, -1)
        substitute = numerix.repeat(faceVertexIDs[numerix.newaxis, 0],
                                    faceVertexIDs.shape[0], axis=0)
        faceVertexIDs = numerix.where(MA.getmaskarray(self.faceVertexIDs),
                                      substitute, faceVertexIDs)
        faceVertexCoords = numerix.take(self.vertexCoords, faceVertexIDs, axis=1)
        faceOrigins = numerix.repeat(faceVertexCoords[:, 0], faceVertexIDs.shape[0], axis=0)
        faceOrigins = numerix.reshape(faceOrigins, MA.shape(faceVertexCoords))
        faceVertexCoords = faceVertexCoords - faceOrigins
        left = list(range(faceVertexIDs.shape[0]))
        right = left[1:] + [left[0]]
        cross = numerix.sum(numerix.cross(faceVertexCoords,
                                          numerix.take(faceVertexCoords, right, 1),
                                          axis=0),
                            1)
        return numerix.sqrtDot(cross, cross) / 2.

    def _calcFaceCenters(self):
        maskedFaceVertexIDs = MA.filled(self.faceVertexIDs, 0)

        faceVertexCoords = numerix.take(self.vertexCoords, maskedFaceVertexIDs, axis=1)

        if MA.getmask(self.faceVertexIDs) is False:
            faceVertexCoordsMask = numerix.zeros(numerix.shape(faceVertexCoords), 'l')
        else:
            faceVertexCoordsMask = \
              numerix.repeat(MA.getmaskarray(self.faceVertexIDs)[numerix.newaxis, ...],
                             self.dim, axis=0)

        faceVertexCoords = MA.array(data=faceVertexCoords, mask=faceVertexCoordsMask)

        return MA.filled(MA.average(faceVertexCoords, axis=1))

    @property
    def _rightHandOrientation(self):
        faceVertexIDs = MA.filled(self.faceVertexIDs, 0)
        faceVertexCoords = numerix.take(self.vertexCoords, faceVertexIDs, axis=1)
        t1 = faceVertexCoords[:, 1,:] - faceVertexCoords[:, 0,:]
        t2 = faceVertexCoords[:, 2,:] - faceVertexCoords[:, 1,:]
        norm = numerix.cross(t1, t2, axis=0)
        ## reordering norm's internal memory for inlining
        norm = norm.copy()
        norm = norm / numerix.sqrtDot(norm, norm)

        faceNormals = -norm

        return 1 - 2 * (numerix.dot(faceNormals, self.cellDistanceVectors) < 0)

    def _calcFaceNormals(self):
        faceVertexIDs = MA.filled(self.faceVertexIDs, 0)
        faceVertexCoords = numerix.take(self.vertexCoords, faceVertexIDs, axis=1)
        t1 = faceVertexCoords[:, 1,:] - faceVertexCoords[:, 0,:]
        t2 = faceVertexCoords[:, 2,:] - faceVertexCoords[:, 1,:]
        norm = numerix.cross(t1, t2, axis=0)
        ## reordering norm's internal memory for inlining
        norm = norm.copy()
        norm = norm / numerix.sqrtDot(norm, norm)

        faceNormals = -norm

        orientation = 1 - 2 * (numerix.dot(faceNormals, self.cellDistanceVectors) < 0)
        return faceNormals * orientation

    def _calcFaceCellToCellNormals(self):
        faceCellCentersUp = numerix.take(self._cellCenters, self.faceCellIDs[1], axis=1)
        faceCellCentersDown = numerix.take(self._cellCenters, self.faceCellIDs[0], axis=1)
        faceCellCentersUp = numerix.where(MA.getmaskarray(faceCellCentersUp),
                                          self._faceCenters,
                                          faceCellCentersUp)

        diff = faceCellCentersDown - faceCellCentersUp
        mag = numerix.sqrt(numerix.sum(diff**2))
        faceCellToCellNormals = diff / numerix.resize(mag, (self.dim, len(mag)))

        orientation = 1 - 2 * (numerix.dot(self.faceNormals, faceCellToCellNormals) < 0)
        return faceCellToCellNormals * orientation

    def _calcOrientedFaceNormals(self):
        return self.faceNormals

    def _calcCellVolumes(self):
        tmp = self._faceCenters[0] * self._faceAreas * self.faceNormals[0]
        tmp = numerix.take(tmp, self.cellFaceIDs) * self._cellToFaceOrientations
        return MA.filled(MA.sum(tmp, 0))

    def _calcCellCenters(self):
        tmp = numerix.take(self._faceCenters, self.cellFaceIDs, axis=1)
        return MA.filled(MA.average(tmp, 1))

    def _calcFaceToCellDistAndVec(self):
        tmp = MA.repeat(self._faceCenters[..., numerix.NewAxis,:], 2, 1)
        # array -= masked_array screws up masking for on numpy 1.1

        tmp = tmp - numerix.take(self._cellCenters, self.faceCellIDs, axis=1)
        cellToFaceDistanceVectors = tmp
        faceToCellDistances = MA.sqrt(MA.sum(tmp * tmp, 0))
        return faceToCellDistances, cellToFaceDistanceVectors

    def _calcCellDistAndVec(self):
        tmp = numerix.take(self._cellCenters, self.faceCellIDs, axis=1)
        tmp = tmp[..., 1,:] - tmp[..., 0,:]
        tmp = MA.filled(MA.where(MA.getmaskarray(tmp), self._cellToFaceDistanceVectors[:, 0], tmp))
        cellDistanceVectors = tmp
        cellDistances = MA.filled(MA.sqrt(MA.sum(tmp * tmp, 0)))
        return cellDistances, cellDistanceVectors

    def _calcFaceTangents(self):
        faceVertexCoord = numerix.array(numerix.take(self.vertexCoords,
                                                     self.faceVertexIDs[0],
                                                     axis=1))
        tmp = self._faceCenters - faceVertexCoord
        faceTangents1 = tmp / numerix.sqrtDot(tmp, tmp)
        tmp = numerix.cross(faceTangents1, self.faceNormals, axis=0)
        faceTangents2 = tmp / numerix.sqrtDot(tmp, tmp)
        return faceTangents1, faceTangents2

    def _calcCellToCellDist(self):
        return numerix.take(self._cellDistances, self.cellFaceIDs)

    def _calcCellAreas(self):
        from fipy.tools.numerix import take
        return take(self._faceAreas, self.cellFaceIDs)

    def _calcCellNormals(self):
        cellNormals = numerix.take(self.faceNormals, self.cellFaceIDs, axis=1)
        cellFaceCellIDs = numerix.take(self.faceCellIDs[0], self.cellFaceIDs)
        cellIDs = numerix.repeat(numerix.arange(self.numberOfCells)[numerix.newaxis, ...],
                                 self._maxFacesPerCell,
                                 axis=0)
        direction = (cellFaceCellIDs == cellIDs) * 2 - 1
        if self._maxFacesPerCell > 0:
            return direction[numerix.newaxis, ...] * cellNormals
        else:
            return cellNormals

    """settable geometry properties"""
    def _getFaceToCellDistances(self):
        return self._internalFaceToCellDistances

    def _setFaceToCellDistances(self, v):
        self._internalFaceToCellDistances = v
        self._setScaledValues()

    _faceToCellDistances = property(_getFaceToCellDistances,
                                    _setFaceToCellDistances)

    def _getCellDistances(self):
        return self._internalCellDistances

    def _setCellDistances(self, v):
        self._internalCellDistances = v
        self._setScaledValues()

    _cellDistances = property(_getCellDistances, _setCellDistances)

    """
    Scaled geometry set and calc
    """

    _scale = {
        'length': 1.,
        'area': 1.,
        'volume': 1.,
    }

    def _setScaledGeometry(self, val):
        """
        Set the scale by length.

        Parameters
        ----------
        val : float
            The new scale length.
        """

        self._scale['length'] = PhysicalField(value=val)

        if self._scale['length'].unit.isDimensionless():
            self._scale['length'] = 1

        self._scale['area'] = self._calcAreaScale()
        self._scale['volume'] = self._calcVolumeScale()
        self._setScaledValues()

    def _setScaledValues(self):
        self._scaledFaceAreas = self._scale['area'] * self._faceAreas
        self._scaledCellVolumes = self._scale['volume'] * self._cellVolumes
        self._scaledCellCenters = self._scale['length'] * self._cellCenters
        self._scaledFaceToCellDistances = self._scale['length'] * self._faceToCellDistances
        self._scaledCellDistances = self._scale['length'] * self._cellDistances
        self._setFaceDependentScaledValues()

    def _setFaceDependentScaledValues(self):
        self._scaledCellToCellDistances = self._scale['length'] * self._cellToCellDistances
        self._areaProjections = self._calcAreaProjections()
        self._orientedAreaProjections = self._calcOrientedAreaProjections()
        self._faceToCellDistanceRatio = self._calcFaceToCellDistanceRatio()
        self._faceAspectRatios = self._calcFaceAspectRatios()

    def _calcAreaScale(self):
        return self.scale['length']**2

    def _calcVolumeScale(self):
        return self.scale['length']**3

    def _calcAreaProjections(self):
        return self.faceNormals * self._faceAreas

    def _calcOrientedAreaProjections(self):
        return self._areaProjections

    def _calcFaceToCellDistanceRatio(self):
        dAP = self._cellDistances
        dFP = self._faceToCellDistances[0]

        return MA.filled(dFP / dAP)

    def _calcFaceAspectRatios(self):
        return self._scaledFaceAreas / self._cellDistances

    def __mul__(self, factor):
        """
        Dilate a `Mesh` by `factor`.

            >>> from fipy.meshes import Grid2D
            >>> baseMesh = Grid2D(dx = 1.0, dy = 1.0, nx = 2, ny = 2)
            >>> print(baseMesh.cellCenters)
            [[ 0.5  1.5  0.5  1.5]
             [ 0.5  0.5  1.5  1.5]]

        The `factor` can be a scalar

            >>> dilatedMesh = baseMesh * 3
            >>> print(dilatedMesh.cellCenters)
            [[ 1.5  4.5  1.5  4.5]
             [ 1.5  1.5  4.5  4.5]]

        or a vector

            >>> dilatedMesh = baseMesh * ((3,), (2,))
            >>> print(dilatedMesh.cellCenters)
            [[ 1.5  4.5  1.5  4.5]
             [ 1.   1.   3.   3. ]]


        but the vector must have the same dimensionality as the `Mesh`

            >>> dilatedMesh = baseMesh * ((3,), (2,), (1,)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            ValueError: shape mismatch: objects cannot be broadcast to a single shape
        """
        newCoords = self.vertexCoords * factor
        newmesh = Mesh(vertexCoords=newCoords,
                       faceVertexIDs=numerix.array(self.faceVertexIDs),
                       cellFaceIDs=numerix.array(self.cellFaceIDs))
        return newmesh

    __rmul__ = __mul__

    @property
    def _concatenableMesh(self):
        return self

    def _translate(self, vector):
        newCoords = self.vertexCoords + vector
        newmesh = Mesh(newCoords, numerix.array(self.faceVertexIDs), numerix.array(self.cellFaceIDs))
        return newmesh

    def _handleFaceConnection(self):
        """
        The `_faceCellToCellNormals` were added to ensure `faceNormals == _faceCellToCellNormals` for periodic grids.

        >>> from fipy import *
        >>> m = PeriodicGrid2DLeftRight(nx=2, ny=2)
        >>> (m.faceNormals == m._faceCellToCellNormals).all()
        True

        """
        self._cellToCellDistances = self._calcCellToCellDist()
        self._faceCellToCellNormals = self._calcFaceCellToCellNormals()
        self._setFaceDependentScaledValues()

    """calculate Topology methods"""

    def _calcFaceCellIDs(self):
        array = MA.array(MA.indices(self.cellFaceIDs.shape, 'l')[1],
                         mask=MA.getmask(self.cellFaceIDs))
        faceCellIDs = MA.zeros((2, self.numberOfFaces), 'l')

        ## Nasty bug: MA.put(arr, ids, values) fills its ids and
        ## values arguments when masked!  This was not the behavior
        ## that was assumed when used below.  It was only working
        ## because the old fill value was 0 and the first element of
        ## the array needed to be 0 since the cell's face was
        ## 0. numerix.put() has been changed to deal with this
        ## properly.

##         MA.put(firstRow, cellFaceIDsFlat[::-1], array[::-1])
##         MA.put(secondRow, cellFaceIDsFlat, array)
        firstRow = faceCellIDs[0]
        secondRow = faceCellIDs[1]

        numerix.put(firstRow, self.cellFaceIDs[::-1, ::-1], array[::-1, ::-1])
        numerix.put(secondRow, self.cellFaceIDs, array)

        mask = ((False,) * self.numberOfFaces, (firstRow == secondRow))
        return MA.sort(MA.array(faceCellIDs, mask = mask),
                                   axis=0)

    """get Topology methods"""

    @property
    def _maxFacesPerCell(self):
        return self.cellFaceIDs.shape[0]

    @property
    def _cellVertexIDs(self):
        ## Get all the vertices from all the faces for each cell
        cellFaceVertices = numerix.take(self.faceVertexIDs, self.cellFaceIDs, axis=1)

        ## get a sorted list of vertices for each cell
        cellVertexIDs = numerix.reshape(cellFaceVertices, (-1, self.numberOfCells))
        cellVertexIDs = MA.sort(cellVertexIDs, axis=0, fill_value=-1)

        cellVertexIDs = MA.sort(MA.concatenate((cellVertexIDs[-1, numerix.newaxis],
                                                MA.masked_where(cellVertexIDs[:-1]
                                                                == cellVertexIDs[1:],
                                                                cellVertexIDs[:-1]))),
                                axis=0, fill_value=-1)

        ## resize the array to remove extra masked values
        if cellVertexIDs.shape[-1] == 0:
            length = 0
        else:
            length = min(numerix.sum(MA.getmaskarray(cellVertexIDs), axis=0))
        return cellVertexIDs[length:][::-1]

#     Below is an ordered version of _getCellVertexIDs()
#     It works for the test case in this file (other than the ordering, obviously)
#     I've left it in as it may be useful when we need ordered vertices for cells
# 
#     def _getOrderedCellVertexIDs(self):
# 
#         ## Get all the vertices from all the faces for each cell
#         from fipy.tools.numerix import take
#         cellFaceVertices = take(self.faceVertexIDs, self.cellFaceIDs)
# 
#         ## get a sorted list of vertices for each cell
#         NCells = self.numberOfCells
#         cellVertexIDs = MA.reshape(cellFaceVertices.flat, (NCells, -1))
#         newmask = MA.getmaskarray(cellVertexIDs).copy()
# 
#         for i in range(len(cellVertexIDs[0]) - 1):
#             for j in range(len(cellVertexIDs[0]))[i + 1:]:
# 
#                 newmask[:,j] = MA.where(newmask[:,j],
#                                         newmask[:,j],
#                                         MA.where(MA.filled(cellVertexIDs)[:,i] == MA.filled(cellVertexIDs)[:,j],
#                                                  1,
#                                                  newmask[:,j]))
# 
# 
#         cellVertexIDs = MA.masked_array(cellVertexIDs, newmask)
# 
#         for i in range(len(cellVertexIDs[0]) - 1):
#             j = i + 1
#             while j < len(cellVertexIDs[0]):
#                 tmp = cellVertexIDs[:]
#                 tmp[:, i] = MA.where(MA.getmaskarray(cellVertexIDs[:,i]),
#                                      MA.where(MA.getmaskarray(cellVertexIDs[:, j]),
#                                               cellVertexIDs[:, i],
#                                               cellVertexIDs[:, j]),
#                                      cellVertexIDs[:, i])
# 
#                 tmp[:, j] = MA.where(MA.getmaskarray(cellVertexIDs[:,i]),
#                                      MA.where(MA.getmaskarray(cellVertexIDs[:, j]),
#                                               cellVertexIDs[:,j],
#                                               cellVertexIDs[:,i]),
#                                      cellVertexIDs[:, j])
# 
#                 cellVertexIDs = tmp[:]
# 
#                 j += 1
# 
# 
#         length = len(cellVertexIDs[0]) - min(numerix.sum(MA.getmaskarray(cellVertexIDs), axis = 1))
#         return cellVertexIDs[:, :length]


    """scaling"""

    def _getNearestCellID(self, points):
        """
        Test cases

           >>> from fipy import *
           >>> m0 = Grid2D(dx=(.1, 1., 10.), dy=(.1, 1., 10.))
           >>> m1 = Grid2D(nx=2, ny=2, dx=5., dy=5.)
           >>> print(m0._getNearestCellID(m1.cellCenters.globalValue))
           [4 5 7 8]

        """
        return numerix.nearest(data=self.cellCenters.globalValue, points=points)

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected.

            >>> dx = 2.
            >>> dy = 1.23456
            >>> dz = 1.e-1

            >>> vertices = numerix.array(((0., 1., 1., 0., 0., 1., 1., 0., 2., 2.),
            ...                           (0., 0., 1., 1., 0., 0., 1., 1., 0., 0.),
            ...                           (0., 0., 0., 0., 1., 1., 1., 1., 0., 1.)))
            >>> vertices *= numerix.array([[dx], [dy], [dz]])
            >>> faces = MA.masked_values(((0, 7, 3, 5, 1, 3, 1, 9, 8, 8),
            ...                           (1, 6, 7, 6, 0, 2, 8, 5, 1, 9),
            ...                           (2, 5, 4, 2, 4, 6, 2, 6, 5, 6),
            ...                           (3, 4, 0, 1, 5, 7, -1, -1, 9, 2)), -1)
            >>> cells = MA.masked_values(((0, 3),
            ...                           (1, 6),
            ...                           (2, 7),
            ...                           (3, 8),
            ...                           (4, 9),
            ...                           (5, -1)), -1)

            >>> mesh = Mesh(vertexCoords=vertices, faceVertexIDs=faces, cellFaceIDs=cells)

            >>> externalFaces = numerix.array((0, 1, 2, 4, 5, 6, 7, 8, 9))
            >>> print(numerix.allequal(externalFaces,
            ...                        numerix.nonzero(mesh.exteriorFaces)))
            1

            >>> internalFaces = numerix.array((3,))
            >>> print(numerix.allequal(internalFaces,
            ...                        numerix.nonzero(mesh.interiorFaces)))
            1

            >>> from fipy.tools.numerix import MA
            >>> faceCellIds = MA.masked_values((( 0,  0,  0, 0,  0,  0,  1,  1,  1,  1),
            ...                                 (-1, -1, -1, 1, -1, -1, -1, -1, -1, -1)), -1)
            >>> numerix.allequal(faceCellIds, mesh.faceCellIDs)
            1

            >>> dxdy = dx * dy
            >>> dxdz = dx * dz
            >>> dydz = dy * dz
            >>> faceAreas = numerix.array((dxdy, dxdy, dydz, dydz, dxdz, dxdz,
            ...                            dxdy/2., dxdy/2., dxdz, numerix.sqrt(dx**2 + dy**2) * dz))
            >>> numerix.allclose(faceAreas, mesh._faceAreas, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceCoords = numerix.take(vertices, MA.filled(faces, 0), axis=1)
            >>> faceCenters = faceCoords[..., 0,:] + faceCoords[..., 1,:] + faceCoords[..., 2,:] + faceCoords[..., 3,:]
            >>> numVex = numerix.array((4., 4., 4., 4., 4., 4., 3., 3., 4., 4.))
            >>> faceCenters /= numVex
            >>> print(numerix.allclose(faceCenters, mesh.faceCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> faceNormals = numerix.array((( 0., 0., -1., 1.,  0., 0.,  0., 0.,  0., dy / numerix.sqrt(dy**2 + dx**2)),
            ...                              ( 0., 0.,  0., 0., -1., 1.,  0., 0., -1., dx / numerix.sqrt(dy**2 + dx**2)),
            ...                              (-1., 1.,  0., 0.,  0., 0., -1., 1.,  0., 0.)))
            >>> numerix.allclose(faceNormals, mesh.faceNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToFaceOrientations = MA.masked_values(((1, -1),
            ...                                            (1, 1),
            ...                                            (1, 1),
            ...                                            (1, 1),
            ...                                            (1, 1),
            ...                                            (1, -2)), -2)
            >>> numerix.allequal(cellToFaceOrientations, mesh._cellToFaceOrientations)
            1

            >>> cellVolumes = numerix.array((dx*dy*dz, dx*dy*dz / 2.))
            >>> numerix.allclose(cellVolumes, mesh.cellVolumes, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellCenters = numerix.array(((dx/2., dx+dx/3.),
            ...                              (dy/2.,    dy/3.),
            ...                              (dz/2.,    dz/2.)))
            >>> print(numerix.allclose(cellCenters, mesh.cellCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> d1 = numerix.sqrt((dx / 3.)**2 + (dy / 6.)**2)
            >>> d2 = numerix.sqrt((dx / 6.)**2 + (dy / 3.)**2)
            >>> d3 = numerix.sqrt((dx / 6.)**2 + (dy / 6.)**2)
            >>> d4 = numerix.sqrt((5 * dx / 6.)**2 + (dy / 6.)**2)
            >>> faceToCellDistances = MA.masked_values(((dz / 2., dz / 2., dx / 2., dx / 2., dy / 2., dy / 2., dz / 2., dz / 2., d2, d3),
            ...                                         (     -1,      -1,      -1,      d1,      -1,      -1,      -1,      -1, -1, -1)), -1)
            >>> print(numerix.allclose(faceToCellDistances, mesh._faceToCellDistances, atol = 1e-10, rtol = 1e-10))
            True

            >>> cellDistances = numerix.array((dz / 2., dz / 2., dx / 2.,
            ...                                d4,
            ...                                dy / 2., dy / 2., dz / 2., dz / 2.,
            ...                                d2,
            ...                                d3))
            >>> numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)
            1

            >>> areaProjections = faceNormals * faceAreas
            >>> numerix.allclose(areaProjections, mesh._areaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> v1 = numerix.take(vertices, numerix.array(faces[0]), axis=1)
            >>> tmp = faceCenters - v1
            >>> tangents1 = tmp / numerix.sqrtDot(tmp, tmp)
            >>> numerix.allclose(tangents1, mesh._faceTangents1, atol = 1e-10, rtol = 1e-10)
            1

            >>> tmp = numerix.cross(tangents1, faceNormals, axis=0)
            >>> tangents2 = tmp / numerix.sqrtDot(tmp, tmp)
            >>> numerix.allclose(tangents2, mesh._faceTangents2, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellToCellIDs = MA.masked_values(((-1,  0),
            ...                                   (-1, -1),
            ...                                   (-1, -1),
            ...                                   ( 1, -1),
            ...                                   (-1, -1),
            ...                                   (-1, -1)), -1)
            >>> numerix.allequal(cellToCellIDs, mesh._cellToCellIDs)
            1

            >>> cellToCellDistances = MA.masked_values(((dz / 2., d4),
            ...                                         (dz / 2., -1),
            ...                                         (dx / 2., -1),
            ...                                         (     d4, -1),
            ...                                         (dy / 2., -1),
            ...                                         (dy / 2., -1)), -1)
            >>> numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)
            1

            >>> interiorCellIDs = numerix.array(())
            >>> numerix.allequal(interiorCellIDs, mesh._interiorCellIDs)
            1

            >>> exteriorCellIDs = numerix.array((0, 1))
            >>> numerix.allequal(exteriorCellIDs, mesh._exteriorCellIDs)
            1

            >>> cellNormals = MA.masked_values((((0, -1),
            ...                                  (0, 0),
            ...                                  (-1, 0),
            ...                                  (1, 0),
            ...                                  (0, dy / numerix.sqrt(dx**2 + dy**2)),
            ...                                  (0, -1000)),
            ...                                 ((0, 0),
            ...                                  (0, 0),
            ...                                  (0, 0),
            ...                                  (0, -1),
            ...                                  (-1, dx / numerix.sqrt(dx**2 + dy**2)),
            ...                                  (1, -1000)),
            ...                                 ((-1, 0),
            ...                                  (1, -1),
            ...                                  (0, 1),
            ...                                  (0, 0),
            ...                                  (0, 0),
            ...                                  (0, -1000))), -1000)
            >>> numerix.allclose(cellNormals, mesh._cellNormals, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellAreaProjections = MA.masked_values((((0, -dy * dz),
            ...                                          (0, 0),
            ...                                          (-dy * dz, 0),
            ...                                          ( dy * dz, 0),
            ...                                          (0, dy * dz),
            ...                                          (0, -1000)),
            ...                                         ((0, 0),
            ...                                          (0, 0),
            ...                                          (0, 0),
            ...                                          (0, -dx * dz),
            ...                                          (-dx * dz, dx * dz),
            ...                                          ( dx * dz, -1000)),
            ...                                         ((-dx * dy, 0),
            ...                                          ( dx * dy, -dx * dy / 2.),
            ...                                          (0, dx * dy / 2.),
            ...                                          (0, 0),
            ...                                          (0, 0),
            ...                                          (0, -1000))), -1000)
            >>> numerix.allclose(cellAreaProjections, mesh._cellAreaProjections, atol = 1e-10, rtol = 1e-10)
            1

            >>> cellVertexIDs = MA.masked_values(((7, 6, 5, 4, 3, 2, 1, 0), (9, 8, 6, 5, 2, 1, -1000, -1000)), -1000)
            >>> cellVertexIDs = MA.masked_values(((7, 9),
            ...                                   (6, 8),
            ...                                   (5, 6),
            ...                                   (4, 5),
            ...                                   (3, 2),
            ...                                   (2, 1),
            ...                                   (1, -1000),
            ...                                   (0, -1000)), -1000)
            >>> numerix.allclose(cellVertexIDs, mesh._cellVertexIDs)
            1


            >>> from fipy.tools import dump
            >>> (f, filename) = dump.write(mesh, extension = '.gz')
            >>> unpickledMesh = dump.read(filename, f)

            >>> print(numerix.allequal(mesh.cellCenters, unpickledMesh.cellCenters))
            True

            >>> dx = 1.
            >>> dy = 1.
            >>> nx = 10
            >>> ny = 2
            >>> from fipy.meshes import Grid2D
            >>> gridMesh = Grid2D(dx, dy, nx, ny)
            >>> from fipy.meshes.tri2D import Tri2D
            >>> triMesh = Tri2D(dx, dy, nx, 1) + [[dx*nx], [0]]
            >>> bigMesh = gridMesh + triMesh
            >>> x, y = bigMesh.cellCenters
            >>> from fipy.variables.cellVariable import CellVariable
            >>> volumes = CellVariable(mesh=bigMesh, value=1.)
            >>> volumes[x > dx * nx] = 0.25
            >>> print(numerix.allclose(bigMesh.cellVolumes, volumes))
            True

            Following test was added due to a bug in adding `UniformGrids`.

            >>> from fipy.meshes.uniformGrid1D import UniformGrid1D
            >>> a = UniformGrid1D(nx=10) + (10,)
            >>> print(a.cellCenters)
            [[ 10.5  11.5  12.5  13.5  14.5  15.5  16.5  17.5  18.5  19.5]]
            >>> b = 10 + UniformGrid1D(nx=10)
            >>> print(b.cellCenters)
            [[ 10.5  11.5  12.5  13.5  14.5  15.5  16.5  17.5  18.5  19.5]]

            >>> c = UniformGrid1D(nx=10) + (UniformGrid1D(nx=10) + 10) # doctest: +SERIAL
            >>> print(numerix.allclose(c.cellCenters[0],
            ...                        [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5, 11.5,
            ...                        12.5, 13.5, 14.5, 15.5, 16.5, 17.5, 18.5, 19.5]))
            ... # doctest: +SERIAL
            True

        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()


